import functools
import inspect
import logging
import sys
from types import CodeType, FunctionType, ModuleType
from typing import Any, Callable, Collection, Dict, List, Optional, Set, Tuple, Type, Union

import numpy as np
import pandas as pd
from pyspark.sql import SparkSession

try:
    import pyspark.pandas as ps
    from pyspark.sql import Column, DataFrame, dataframe, types
    from pyspark.sql.functions import column, lit, pandas_udf, udf
except ImportError:
    raise NotImplementedError("Pyspark is not installed.")

from hamilton import base, htypes, node
from hamilton.execution import graph_functions
from hamilton.function_modifiers import base as fm_base
from hamilton.function_modifiers import subdag
from hamilton.function_modifiers.recursive import assign_namespace
from hamilton.htypes import custom_subclass_check
from hamilton.lifecycle import base as lifecycle_base

logger = logging.getLogger(__name__)


class KoalasDataFrameResult(base.ResultMixin):
    """Mixin for building a koalas dataframe from the result"""

    @staticmethod
    def build_result(**outputs: Dict[str, Any]) -> ps.DataFrame:
        """Right now this class is just used for signaling the return type."""
        pass


class SparkKoalasGraphAdapter(base.HamiltonGraphAdapter, base.ResultMixin):
    """Class representing what's required to make Hamilton run on Spark with Koalas, i.e. Pandas on Spark.

    This walks the graph and translates it to run onto `Apache Spark <https://spark.apache.org/">`__ \
    using the \
    `Pandas API on Spark <https://spark.apache.org/docs/latest/api/python/user_guide/pandas_on_spark/index.html>`__

    Use `pip install sf-hamilton[spark]` to get the dependencies required to run this.

    Currently, this class assumes you're running SPARK 3.2+. You'd generally use this if you have an existing spark \
    cluster running in your workplace, and you want to scale to very large data set sizes.

    Some tips on koalas (before it was merged into spark 3.2):

     - https://databricks.com/blog/2020/03/31/10-minutes-from-pandas-to-koalas-on-apache-spark.html
     - https://spark.apache.org/docs/latest/api/python/user_guide/pandas_on_spark/index.html

    Spark is a more heavyweight choice to scale computation for Hamilton graphs creating a Pandas Dataframe.

    Notes on scaling:
    -----------------
      - Multi-core on single machine ✅ (if you setup Spark locally to do so)
      - Distributed computation on a Spark cluster ✅
      - Scales to any size of data as permitted by Spark ✅

    Function return object types supported:
    ---------------------------------------
      - ⛔ Not generic. This does not work for every Hamilton graph.
      - ✅ Currently we're targeting this at Pandas/Koalas types [dataframes, series].

    Pandas?
    -------
      - ✅ Koalas on Spark 3.2+ implements a good subset of the pandas API. Keep it simple and you should be good to go!

    CAVEATS
    -------
      - Serialization costs can outweigh the benefits of parallelism, so you should benchmark your code to see if it's\
      worth it.

    DISCLAIMER -- this class is experimental, so signature changes are a possibility!
    """

    def __init__(self, spark_session, result_builder: base.ResultMixin, spine_column: str):
        """Constructor

        You only have the ability to return either a Pandas on Spark Dataframe or a Pandas Dataframe. To do that you \
        either use the stock \
        `base.PandasDataFrameResult <https://github.com/dagworks-inc/hamilton/blob/main/hamilton/base.py#L39>`__ class,\
         or you use `h_spark.KoalasDataframeResult <https://github.com/dagworks-inc/hamilton/blob/main/hamilton/experimental/h_spark.py#L16>`__.

        :param spark_session: the spark session to use.
        :param result_builder: the function to build the result -- currently on Pandas and Koalas are "supported".
        :param spine_column: the column we should use first as the spine and then subsequently join against.
        """
        self.spark_session = spark_session
        if not (
            isinstance(result_builder, base.PandasDataFrameResult)
            or isinstance(result_builder, KoalasDataFrameResult)
            or isinstance(result_builder, base.DictResult)
        ):
            raise ValueError(
                "SparkKoalasGraphAdapter only supports returning:"
                ' a "pandas" DF at the moment, a "koalas" DF at the moment, or a "dict" of results.'
            )
        self.result_builder = result_builder
        self.spine_column = spine_column

    @staticmethod
    def check_input_type(node_type: Type, input_value: Any) -> bool:
        """Function to equate an input value, with expected node type.

        We need this to equate pandas and koalas objects/types.

        :param node_type: the declared node type
        :param input_value: the actual input value
        :return: whether this is okay, or not.
        """
        # TODO: flesh this out more
        if (node_type == pd.Series or node_type == ps.Series) and (
            isinstance(input_value, ps.DataFrame) or isinstance(input_value, ps.Series)
        ):
            return True
        elif node_type == np.array and isinstance(input_value, dataframe.DataFrame):
            return True

        return htypes.check_input_type(node_type, input_value)

    @staticmethod
    def check_node_type_equivalence(node_type: Type, input_type: Type) -> bool:
        """Function to help equate pandas with koalas types.

        :param node_type: the declared node type.
        :param input_type: the type of what we want to pass into it.
        :return: whether this is okay, or not.
        """
        if node_type == ps.Series and input_type == pd.Series:
            return True
        elif node_type == pd.Series and input_type == ps.Series:
            return True
        elif node_type == ps.DataFrame and input_type == pd.DataFrame:
            return True
        elif node_type == pd.DataFrame and input_type == ps.DataFrame:
            return True
        return node_type == input_type

    def execute_node(self, node: node.Node, kwargs: Dict[str, Any]) -> Any:
        """Function that is called as we walk the graph to determine how to execute a hamilton function.

        :param node: the node from the graph.
        :param kwargs: the arguments that should be passed to it.
        :return: returns a koalas column
        """
        return node.callable(**kwargs)

    def build_result(self, **outputs: Dict[str, Any]) -> Union[pd.DataFrame, ps.DataFrame, dict]:
        if isinstance(self.result_builder, base.DictResult):
            return self.result_builder.build_result(**outputs)
        # we don't use the actual function for building right now, we use this hacky equivalent
        df = ps.DataFrame(outputs[self.spine_column])
        for k, v in outputs.items():
            logger.info(f"Got column {k}, with type [{type(v)}].")
            df[k] = v
        if isinstance(self.result_builder, base.PandasDataFrameResult):
            return df.to_pandas()
        else:
            return df


def numpy_to_spark_type(numpy_type: Type) -> types.DataType:
    """Function to convert a numpy type to a Spark type.

    :param numpy_type: the numpy type to convert.
    :return: the Spark type.
    :raise: ValueError if the type is not supported.
    """
    if (
        numpy_type == np.int8
        or numpy_type == np.int16
        or numpy_type == np.int32
        or numpy_type == np.int64
    ):
        return types.IntegerType()
    elif numpy_type == np.float16 or numpy_type == np.float32 or numpy_type == np.float64:
        return types.FloatType()
    elif numpy_type == np.bool_:
        return types.BooleanType()
    elif numpy_type == np.unicode_ or numpy_type == np.string_:
        return types.StringType()
    elif numpy_type == np.bytes_:
        return types.BinaryType()
    else:
        raise ValueError("Unsupported NumPy type: " + str(numpy_type))


def python_to_spark_type(python_type: Type[Union[int, float, bool, str, bytes]]) -> types.DataType:
    """Function to convert a Python type to a Spark type.

    :param python_type: the Python type to convert.
    :return: the Spark type.
    :raise: ValueError if the type is not supported.
    """
    if python_type == int:
        return types.IntegerType()
    elif python_type == float:
        return types.FloatType()
    elif python_type == bool:
        return types.BooleanType()
    elif python_type == str:
        return types.StringType()
    elif python_type == bytes:
        return types.BinaryType()
    else:
        raise ValueError("Unsupported Python type: " + str(python_type))


if sys.version_info < (3, 9):
    _list = (List[int], List[float], List[bool], List[str], List[bytes])
else:
    _list = (list[int], list[float], list[bool], list[str], list[bytes])


def get_spark_type(return_type: Any) -> types.DataType:
    if return_type in (int, float, bool, str, bytes):
        return python_to_spark_type(return_type)
    elif return_type in _list:
        return types.ArrayType(python_to_spark_type(return_type.__args__[0]))
    elif hasattr(return_type, "__module__") and getattr(return_type, "__module__") == "numpy":
        return numpy_to_spark_type(return_type)
    else:
        raise ValueError(
            f"Currently unsupported return type {return_type}. "
            f"Please create an issue or PR to add support for this type."
        )


def _get_pandas_annotations(node_: node.Node, bound_parameters: Dict[str, Any]) -> Dict[str, bool]:
    """Given a function, return a dictionary of the parameters that are annotated as pandas series.

    :param hamilton_udf: the function to check.
    :return: dictionary of parameter names to boolean indicating if they are pandas series.
    """

    def _get_type_from_annotation(annotation: Any) -> Any:
        """Gets the type from the annotation if there is one."""
        actual_type, extras = htypes.get_type_information(annotation)
        return actual_type

    return {
        name: _get_type_from_annotation(type_) == pd.Series
        for name, (type_, dep_type) in node_.input_types.items()
        if name not in bound_parameters and dep_type == node.DependencyType.REQUIRED
    }


def _determine_parameters_to_bind(
    actual_kwargs: dict,
    df_columns: Set[str],
    node_input_types: Dict[str, Tuple],
    node_name: str,
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Function that we use to bind inputs to the function, or determine we should pull them from the dataframe.

    It does two things:

    1. If the parameter name matches a column name in the dataframe, create a pyspark column object for it.
    2. If the parameter name matches a key in the input dictionary, and the value is not a dataframe,\
    bind it to the function.

    :param actual_kwargs: the input dictionary of arguments for the function.
    :param df_columns: the set of column names in the dataframe.
    :param node_input_types: the input types of the function.
    :param node_name: name of the node/function.
    :return: a tuple of the params that come from the dataframe and the parameters to bind.
    """
    params_from_df = {}
    bind_parameters = {}
    for input_name, (type_, dep_type) in node_input_types.items():
        if input_name in df_columns:
            params_from_df[input_name] = column(input_name)
        elif input_name in actual_kwargs and not isinstance(actual_kwargs[input_name], DataFrame):
            bind_parameters[input_name] = actual_kwargs[input_name]
        elif dep_type == node.DependencyType.REQUIRED:
            raise ValueError(
                f"Cannot satisfy {node_name} with input types {node_input_types} against a "
                f"dataframe with "
                f"columns {df_columns} and input kwargs {actual_kwargs}."
            )
    return params_from_df, bind_parameters


def _inspect_kwargs(kwargs: Dict[str, Any]) -> Tuple[DataFrame, Dict[str, Any]]:
    """Inspects kwargs, removes any dataframes, and returns the (presumed single) dataframe, with remaining kwargs.

    :param kwargs: the inputs to the function.
    :return: tuple of the dataframe and the remaining non-dataframe kwargs.
    """
    df = None
    actual_kwargs = {}
    for kwarg_key, kwarg_value in kwargs.items():
        if isinstance(kwarg_value, DataFrame):
            if df is None:
                df = kwarg_value
        else:
            actual_kwargs[kwarg_key] = kwarg_value
    return df, actual_kwargs


def _format_pandas_udf(func_name: str, ordered_params: List[str]) -> str:
    formatting_params = {
        "name": func_name,
        "return_type": "pd.Series",
        "params": ", ".join([f"{param}: pd.Series" for param in ordered_params]),
        "param_call": ", ".join([f"{param}={param}" for param in ordered_params]),
    }
    func_string = """
def {name}({params}) -> {return_type}:
    return partial_fn({param_call})
""".format(**formatting_params)
    return func_string


def _format_udf(func_name: str, ordered_params: List[str]) -> str:
    formatting_params = {
        "name": func_name,
        "params": ", ".join(ordered_params),
        "param_call": ", ".join([f"{param}={param}" for param in ordered_params]),
    }
    func_string = """
def {name}({params}):
    return partial_fn({param_call})
""".format(**formatting_params)
    return func_string


def _fabricate_spark_function(
    node_: node.Node,
    params_to_bind: Dict[str, Any],
    params_from_df: Dict[str, Any],
    pandas_udf: bool,
) -> FunctionType:
    """Fabricates a spark compatible UDF. We have to do this as we don't actually have a funtion
    with annotations to use, as its lambdas passed around by decorators. We may consider pushing
    this upstreams so that everything can generate its own function, but for now this is the
    easiest way to do it.

    The rules are different for pandas series and regular UDFs.
    Pandas series have to:
    - be Decorated with pandas_udf
    - Have a return type of a pandas series
    - Have a pandas series as the only input types
    Regular UDFs have to:
    - Have no annotations at all

    See https://spark.apache.org/docs/3.1.3/api/python/reference/api/pyspark.sql.functions.udf.html
    and https://spark.apache.org/docs/3.1.3/api/python/reference/api/pyspark.sql.functions.pandas_udf.html

    :param node_: Node to place in a spark function
    :param params_to_bind: Parameters to bind to the function -- these won't go into the UDF
    :param params_from_df: Parameters to retrieve from the dataframe
    :return: A function that can be used in a spark UDF
    """
    partial_fn = functools.partial(node_.callable, **params_to_bind)
    ordered_params = sorted(params_from_df)
    func_name = node_.name.replace(".", "_")
    if pandas_udf:
        func_string = _format_pandas_udf(func_name, ordered_params)
    else:
        func_string = _format_udf(func_name, ordered_params)
    module_code = compile(func_string, "<string>", "exec")
    func_code = [c for c in module_code.co_consts if isinstance(c, CodeType)][0]
    return FunctionType(func_code, {**globals(), **{"partial_fn": partial_fn}}, func_name)


def _lambda_udf(df: DataFrame, node_: node.Node, actual_kwargs: Dict[str, Any]) -> DataFrame:
    """Function to create a lambda UDF for a function.

    This functions does the following:

    1. Determines whether we can bind any arguments to the function, e.g. primitives.
    2. Determines what type of UDF it is, regular or Pandas, and processes the function accordingly.
    3. Determines the return type of the UDF.
    4. Creates the UDF and applies it to the dataframe.

    :param df: the spark dataframe to apply UDFs to.
    :param node_: the node representing the function.
    :param hamilton_udf: the function to apply.
    :param actual_kwargs: the actual arguments to the function.
    :return: the dataframe with one more column representing the result of the UDF.
    """
    params_from_df, params_to_bind = _determine_parameters_to_bind(
        actual_kwargs, set(df.columns), node_.input_types, node_.name
    )
    pandas_annotation = _get_pandas_annotations(node_, params_to_bind)
    if any(pandas_annotation.values()) and not all(pandas_annotation.values()):
        raise ValueError(
            f"Currently unsupported function for {node_.name} with function signature:\n{node_.input_types}."
        )
    elif all(pandas_annotation.values()) and len(pandas_annotation.values()) > 0:
        hamilton_udf = _fabricate_spark_function(node_, params_to_bind, params_from_df, True)
        # pull from annotation here instead of tag.
        base_type, type_args = htypes.get_type_information(node_.type)
        logger.debug("PandasUDF: %s, %s, %s", node_.name, base_type, type_args)
        if not type_args:
            raise ValueError(
                f"{node_.name} needs to be annotated with htypes.column[pd.Series, TYPE], "
                f"where TYPE could be the string name of the python type, or the python type itself."
            )
        type_arg = type_args[0]
        if isinstance(type_arg, str):
            spark_return_type = type_arg  # spark will handle converting it.
        else:
            spark_return_type = get_spark_type(type_arg)
        spark_udf = pandas_udf(hamilton_udf, spark_return_type)
    else:
        hamilton_udf = _fabricate_spark_function(node_, params_to_bind, params_from_df, False)
        logger.debug("RegularUDF: %s, %s", node_.name, node_.type)
        spark_return_type = get_spark_type(node_.type)
        spark_udf = udf(hamilton_udf, spark_return_type)
    out = df.withColumn(
        node_.name,
        spark_udf(*[_value for _name, _value in sorted(params_from_df.items())]),
    )
    return out


class PySparkUDFGraphAdapter(base.SimplePythonDataFrameGraphAdapter):
    """UDF graph adapter for PySpark.

    This graph adapter enables one to write Hamilton functions that can be executed as UDFs in PySpark.

    Core to this is the mapping of function arguments to Spark columns available in the passed in dataframe.

    This adapter currently supports:

    - regular UDFs, these are executed in a row based fashion.
    - and a single variant of Pandas UDFs: func(series+) -> series
    - can also run regular Hamilton functions, which will execute spark driver side.

    DISCLAIMER -- this class is experimental, so signature changes are a possibility!
    """

    def __init__(self):
        self.df_object = None
        self.original_schema = []
        self.call_count = 0

    @staticmethod
    def check_input_type(node_type: Type, input_value: Any) -> bool:
        """If the input is a pyspark dataframe, skip, else delegate the check."""
        if isinstance(input_value, DataFrame):
            return True
        return htypes.check_input_type(node_type, input_value)

    @staticmethod
    def check_node_type_equivalence(node_type: Type, input_type: Type) -> bool:
        """Checks for the htype.column annotation and deals with it."""
        # Good Cases:
        # [pd.Series, int] -> [pd.Series, int]
        # pd.series -> pd.series
        # [pd.Series, int] -> int
        node_base_type, node_annotations = htypes.get_type_information(node_type)
        input_base_type, input_annotations = htypes.get_type_information(input_type)
        exact_match = node_type == input_type
        series_to_series = node_base_type == input_base_type
        if node_annotations:
            series_to_primitive = node_annotations[0] == input_base_type
        else:
            series_to_primitive = False
        return exact_match or series_to_series or series_to_primitive

    def execute_node(self, node: node.Node, kwargs: Dict[str, Any]) -> Any:
        """Given a node to execute, process it and apply a UDF if applicable.

        :param node: the node we're processing.
        :param kwargs: the inputs to the function.
        :return: the result of the function.
        """
        self.call_count += 1
        logger.debug("%s, %s", self.call_count, self.df_object)
        # get dataframe object out of kwargs
        df, actual_kwargs = _inspect_kwargs(kwargs)
        if df is None:  # there were no dataframes passed in. So regular function call.
            return node.callable(**actual_kwargs)
        if self.df_object is None:
            self.df_object = df  # this is done only once.
            self.original_schema = list(df.columns)
        logger.debug("%s, %s", self.call_count, self.df_object)
        logger.debug("%s, Before, %s", node.name, self.df_object.columns)
        schema_length = len(df.schema)
        df = _lambda_udf(self.df_object, node, actual_kwargs)
        assert node.name in df.columns, f"Error {node.name} not in {df.columns}"
        delta = len(df.schema) - schema_length
        if delta == 0:
            raise ValueError(
                f"UDF {node.name} did not add any columns to the dataframe. "
                f"Does it already exist in the dataframe?"
            )
        self.df_object = df
        logger.debug("%s, After, %s", node.name, df.columns)
        return df

    def build_result(self, **outputs: Dict[str, Any]) -> DataFrame:
        """Builds the result and brings it back to this running process.

        :param outputs: the dictionary of key -> Union[ray object reference | value]
        :return: The type of object returned by self.result_builder.
        """
        df: DataFrame = self.df_object
        output_schema = self.original_schema
        # what's in the dataframe:
        for output_name, output_value in outputs.items():
            if output_name not in output_schema:
                output_schema.append(output_name)
            if output_name in df.columns:
                continue
            else:
                df = df.withColumn(output_name, lit(output_value))
        # original schema + new columns should be the order.
        # if someone requests a column that is in the original schema we won't duplicate it.
        result = df.select(*[column(col_name) for col_name in output_schema])
        # clear state out
        self.df_object = None
        self.original_schema = []
        return result


def sparkify_node_with_udf(
    node_: node.Node,
    linear_df_dependency_name: str,
    base_df_dependency_name: str,
    base_df_dependency_param: Optional[str],
    dependent_columns_in_group: Set[str],
    dependent_columns_from_dataframe: Set[str],
) -> node.Node:
    """ """
    """Turns a node into a spark node. This does the following:
    1. Makes it take the prior dataframe output as a dependency, in
       conjunction to its current dependencies. This is so we can represent
       the "logical" plan (the UDF-dependencies) as well as
       the "physical plan" (linear, df operations)
    2. Adjusts the function to apply the specified UDF on the
       dataframe, ignoring all inputs in column_dependencies
       (which are only there to demonstrate lineage/make the DAG representative)
    3. Returns the resulting pyspark dataframe for downstream functions to use


    :param node_: Node we're sparkifying
    :param linear_df_dependency_name: Name of the linearly passed along dataframe dependency
    :param base_df_dependency_name: Name of the base (parent) dataframe dependency.
        this is only used if dependent_columns_from_dataframe is not empty
    :param base_df_dendency_param: Name of the base (parent) dataframe dependency parameter, as known
        by the node. This is only used if `pass_dataframe_as` is provided, which means that
        dependent_columns_from_dataframe is empty.
    :param dependent_columns_in_group: Columns on which this depends in the with_columns
    :param dependent_columns_from_dataframe:  Columns on which this depends in the
        base (parent) dataframe that the with_columns is operating on
    :return:

    """

    def new_callable(
        __linear_df_dependency_name: str = linear_df_dependency_name,
        __base_df_dependency_name: str = base_df_dependency_name,
        __dependent_columns_in_group: Set[str] = dependent_columns_in_group,
        __dependent_columns_from_dataframe: Set[str] = dependent_columns_from_dataframe,
        __base_df_dependency_param: str = base_df_dependency_param,
        __node: node.Node = node_,
        **kwargs,
    ) -> ps.DataFrame:
        """This is the new function that the node will call.
        Note that this applies the hamilton UDF with *just* the input dataframe dependency,
        ignoring the rest."""
        # gather the dataframe from the kwargs
        df = kwargs[__linear_df_dependency_name]
        kwargs = {
            k: v
            for k, v in kwargs.items()
            if k not in __dependent_columns_from_dataframe
            and k not in __dependent_columns_in_group
            and k != __linear_df_dependency_name
            and k != __base_df_dependency_name
        }
        return _lambda_udf(df, node_, kwargs)

    # Just extract the dependeency type
    # TODO -- add something as a "logical" or "placeholder" dependency
    new_input_types = {
        # copy over the old ones
        **{
            dep: value
            for dep, value in node_.input_types.items()
            if dep not in dependent_columns_from_dataframe
        },
        # add the new one (from the previous)
        linear_df_dependency_name: (DataFrame, node.DependencyType.REQUIRED),
        # Then add all the others
        # Note this might clobber the linear_df_dependency_name, but they'll be the same type
        # If we have "logical" dependencies we'll want to be careful about the type
        **{
            dep: (DataFrame, node.DependencyType.REQUIRED)
            for dep, _ in node_.input_types.items()
            if dep in dependent_columns_in_group
        },
    }

    if base_df_dependency_param is not None and base_df_dependency_name in node_.input_types:
        # In this case we want to add a dependency for visualization/lineage
        new_input_types[base_df_dependency_name] = (
            DataFrame,
            node.DependencyType.REQUIRED,
        )
    if len(dependent_columns_from_dataframe) > 0:
        new_input_types[base_df_dependency_name] = (
            DataFrame,
            node.DependencyType.REQUIRED,
        )
    return node_.copy_with(callabl=new_callable, input_types=new_input_types, typ=DataFrame)


def derive_dataframe_parameter(
    param_types: Dict[str, Type], requested_parameter: str, location_name: Callable
) -> str:
    dataframe_parameters = {
        param for param, val in param_types.items() if custom_subclass_check(val, DataFrame)
    }
    if requested_parameter is not None:
        if requested_parameter not in dataframe_parameters:
            raise ValueError(
                f"Requested parameter {requested_parameter} not found in " f"{location_name}"
            )
        return requested_parameter
    if len(dataframe_parameters) == 0:
        raise ValueError(
            f"No dataframe parameters found in: {location_name}. "
            f"Received parameters: {param_types}. "
            f"@with_columns must inject a dataframe parameter into the function."
        )
    elif len(dataframe_parameters) > 1:
        raise ValueError(
            f"More than one dataframe parameter found in function: {location_name}. Please "
            f"specify the desired one with the 'dataframe' parameter in @with_columns"
        )
    assert len(dataframe_parameters) == 1
    return list(dataframe_parameters)[0]


def derive_dataframe_parameter_from_fn(fn: Callable, requested_parameter: str = None) -> str:
    """Utility function to grab a pyspark dataframe parameter from a function.
    Note if one is supplied it'll look for that. If none is, it will look to ensure
    that there is only one dataframe parameter in the function.

    :param fn: Function to grab the dataframe parameter from
    :param requested_parameter: If supplied, the name of the parameter to grab
    :return: The name of the dataframe parameter
    :raises ValueError: If no datframe parameter is supplied:
    - if no dataframe parameter is found, or if more than one is found
    if a requested parameter is supplied:
    - if the requested parameter is not found
    """
    sig = inspect.signature(fn)
    parameters_with_types = {param.name: param.annotation for param in sig.parameters.values()}
    return derive_dataframe_parameter(parameters_with_types, requested_parameter, fn.__qualname__)


def _derive_first_dataframe_parameter_from_fn(fn: Callable) -> str:
    """Utility function to derive the first parameter from a function and assert
    that it is annotated with a pyspark dataframe.

    :param fn:
    :return:
    """
    sig = inspect.signature(fn)
    params = list(sig.parameters.items())
    if len(params) == 0:
        raise ValueError(
            f"Function {fn.__qualname__} has no parameters, but was "
            f"decorated with with_columns. with_columns requires the first "
            f"parameter to be a dataframe so we know how to wire dependencies."
        )
    first_param_name, first_param_value = params[0]
    if not custom_subclass_check(first_param_value.annotation, DataFrame):
        raise ValueError(
            f"Function {fn.__qualname__} has a first parameter {first_param_name} "
            f"that is not a pyspark dataframe. Instead got: {first_param_value.annotation}."
            f"with_columns requires the first "
            f"parameter to be a dataframe so we know how to wire dependencies."
        )
    return first_param_name


def derive_dataframe_parameter_from_node(node_: node.Node, requested_parameter: str = None) -> str:
    """Derives the only/requested dataframe parameter from a node.

    :param node_:
    :param requested_parameter:
    :return:
    """
    types_ = {key: value[0] for key, value in node_.input_types.items()}
    originating_function_name = (
        node_.originating_functions[-1] if node_.originating_functions is not None else node_.name
    )
    return derive_dataframe_parameter(types_, requested_parameter, originating_function_name)


def prune_nodes(nodes: List[node.Node], select: Optional[List[str]] = None) -> List[node.Node]:
    """Prunes the nodes to only include those upstream from the select columns.
    Conducts a depth-first search using the nodes `input_types` field.

    If select is None, we just assume all nodes should be included.

    :param nodes: Full set of nodes
    :param select: Columns to select
    :return:  Pruned set of nodes
    """
    if select is None:
        return nodes

    node_name_map = {node_.name: node_ for node_ in nodes}
    seen_nodes = set(select)
    stack = list({node_name_map[col] for col in select if col in node_name_map})
    output = []
    while len(stack) > 0:
        node_ = stack.pop()
        output.append(node_)
        for dep in node_.input_types:
            if dep not in seen_nodes and dep in node_name_map:
                dep_node = node_name_map[dep]
                stack.append(dep_node)
            seen_nodes.add(dep)
    return output


class require_columns(fm_base.NodeTransformer):
    """Decorator for spark that allows for the specification of columns to transform.
    These are columns within a specific node in a decorator, enabling the user to make use of pyspark
    transformations inside a with_columns group. Note that this will have no impact if it is not
    decorating a node inside `with_columns`.

    Note that this currently does not work with other decorators, but it definitely could.
    """

    TRANSFORM_TARGET_TAG = "hamilton.spark.target"
    TRANSFORM_COLUMNS_TAG = "hamilton.spark.columns"

    def __init__(self, *columns: str):
        super(require_columns, self).__init__(target=None)
        self._columns = columns

    def transform_node(
        self, node_: node.Node, config: Dict[str, Any], fn: Callable
    ) -> Collection[node.Node]:
        """Generates nodes for the `@require_columns` decorator.

        This does two things, but does not fully prepare the node:
        1. It adds the columns as dependencies to the node
        2. Adds tags with relevant metadata for later use

        Note that, at this point, we don't actually know which columns will come from the
        base dataframe, and which will come from the upstream nodes. This is handled in the
        `with_columns` decorator, so for now, we need to give it enough information to topologically
        sort/assign dependencies.

        :param node_: Node to transform
        :param config: Configuration to use (unused here)
        :return:
        """
        param = derive_dataframe_parameter_from_node(node_)

        # This allows for injection of any extra parameters
        def new_callable(__input_types=node_.input_types, **kwargs):
            return node_.callable(
                **{key: value for key, value in kwargs.items() if key in __input_types}
            )

        additional_input_types = {
            param: (DataFrame, node.DependencyType.REQUIRED)
            for param in self._columns
            if param not in node_.input_types
        }
        node_out = node_.copy_with(
            input_types={**node_.input_types, **additional_input_types},
            callabl=new_callable,
            tags={
                require_columns.TRANSFORM_TARGET_TAG: param,
                require_columns.TRANSFORM_COLUMNS_TAG: self._columns,
            },
        )
        # if it returns a column, we just turn it into a withColumn expression
        if custom_subclass_check(node_.type, Column):

            def transform_output(output: Column, kwargs: Dict[str, Any]) -> DataFrame:
                return kwargs[param].withColumn(node_.name, output)

            node_out = node_out.transform_output(transform_output, DataFrame)
        return [node_out]

    def validate(self, fn: Callable):
        """Validates on the function, even though it operates on nodes. We can always loosen
        this, but for now it should help the code stay readable.

        :param fn: Function this is decorating
        :return:
        """

        _derive_first_dataframe_parameter_from_fn(fn)

    @staticmethod
    def _extract_dataframe_params(node_: node.Node) -> List[str]:
        """Extracts the dataframe parameters from a node.

        :param node_: Node to extract from
        :return: List of dataframe parameters
        """
        return [
            key
            for key, value in node_.input_types.items()
            if custom_subclass_check(value[0], DataFrame)
        ]

    @staticmethod
    def is_default_pyspark_udf(node_: node.Node) -> bool:
        """Tells if a node is, by default, a pyspark UDF. This means:
        1. It has a single dataframe parameter
        2. That parameter name determines an upstream column name

        :param node_: Node to check
        :return: True if it functions as a default pyspark UDF, false otherwise
        """
        df_columns = require_columns._extract_dataframe_params(node_)
        return len(df_columns) == 1

    @staticmethod
    def is_decorated_pyspark_udf(node_: node.Node):
        """Tells if this is a decorated pyspark UDF. This means it has been
        decorated by the `@transforms` decorator.

        :return: True if it can be run as part of a group, false otherwise
        """
        if "hamilton.spark.columns" in node_.tags and "hamilton.spark.target" in node_.tags:
            return True
        return False

    @staticmethod
    def sparkify_node(
        node_: node.Node,
        linear_df_dependency_name: str,
        base_df_dependency_name: str,
        base_df_param_name: Optional[str],
        dependent_columns_from_upstream: Set[str],
        dependent_columns_from_dataframe: Set[str],
    ) -> node.Node:
        """Transforms a pyspark node into a node that can be run as part of a `with_columns` group.
        This is only for non-UDF nodes that have already been transformed by `@transforms`.

        :param node_: Node to transform
        :param linear_df_dependency_name: Dependency on continaully modified dataframe (this will enable us
        :param base_df_dependency_name:
        :param dependent_columns_in_group:
        :param dependent_columns_from_dataframe:
        :return: The final node with correct dependencies
        """
        transformation_target = node_.tags.get(require_columns.TRANSFORM_TARGET_TAG)

        # Note that the following does not use the reassign_columns function as we have
        # special knowledge of the function -- E.G. that it doesn't need all the parameters
        # we choose to pass it. Thus we can just make sure that we pass it the right one,
        # and not worry about value-clashes in reassigning names (as there are all sorts of
        # edge cases around the parameter name to be transformed).

        # We have only a few dependencies we truly need
        # These are the linear_df_dependency_name (the dataframe that is being modified)
        # as well as any non-dataframe arguments (E.G. the ones that aren't about to be added
        # Note that the node comes with logical dependencies already, so we filter them out
        def new_callable(__callable=node_.callable, **kwargs) -> Any:
            new_kwargs = kwargs.copy()
            new_kwargs[transformation_target] = kwargs[linear_df_dependency_name]
            return __callable(**new_kwargs)

        # We start off with everything except the transformation target, as we're
        # going to use the linear dependency for that (see the callable above)
        new_input_types = {
            key: value
            for key, value in node_.input_types.items()
            if key != transformation_target and key not in dependent_columns_from_dataframe
        }
        # Thus we put that linear dependency in
        new_input_types[linear_df_dependency_name] = (
            DataFrame,
            node.DependencyType.REQUIRED,
        )
        # Then we go through all "logical" dependencies -- columns we want to add to make lineage
        # look nice
        for item in dependent_columns_from_upstream:
            new_input_types[item] = (DataFrame, node.DependencyType.REQUIRED)

        # Then we see if we're trying to transform the base dataframe
        # This means we're not referring to it as a column, and only happens with the
        # `pass_dataframe_as` argument (which means the base_df_param_name is not None)
        if transformation_target == base_df_param_name:
            new_input_types[base_df_dependency_name] = (
                DataFrame,
                node.DependencyType.REQUIRED,
            )
        # Finally we create the new node and return it
        node_ = node_.copy_with(callabl=new_callable, input_types=new_input_types)
        return node_


def _identify_upstream_dataframe_nodes(nodes: List[node.Node]) -> List[str]:
    """Gives the upstream dataframe name. This is the only ps.DataFrame parameter not
    produced from within the subdag.

    :param nodes: Nodes in the subdag
    :return: The name of the upstream dataframe
    """
    node_names = {node_.name for node_ in nodes}
    df_deps = set()

    for node_ in nodes:
        # In this case its a df node that is a linear dependency, so we don't count it
        # Instead we count the columns it wants, as we have not yet created them TODO --
        # consider moving this validation afterwards so we don't have to do this check
        df_dependencies = node_.tags.get(
            require_columns.TRANSFORM_COLUMNS_TAG,
            [
                dep
                for dep, (type_, _) in node_.input_types.items()
                if custom_subclass_check(type_, DataFrame)
            ],
        )
        for dependency in df_dependencies:
            if dependency not in node_names:
                df_deps.add(dependency)
    return list(df_deps)


class with_columns(fm_base.NodeCreator):
    def __init__(
        self,
        *load_from: Union[Callable, ModuleType],
        columns_to_pass: List[str] = None,
        pass_dataframe_as: str = None,
        select: List[str] = None,
        namespace: str = None,
        mode: str = "append",
        config_required: List[str] = None,
    ):
        """Initializes a with_columns decorator for spark. This allows you to efficiently run
         groups of map operations on a dataframe, represented as pandas/primitives UDFs. This
         effectively "linearizes" compute -- meaning that a DAG of map operations can be run
         as a set of `.withColumn` operations on a single dataframe -- ensuring that you don't have
         to do a complex `extract` then `join` process on spark, which can be inefficient.

         Here's an example of calling it -- if you've seen `@subdag`, you should be familiar with
         the concepts:

         .. code-block:: python

             # my_module.py
             def a(a_from_df: pd.Series) -> pd.Series:
                 return _process(a)

             def b(b_from_df: pd.Series) -> pd.Series:
                 return _process(b)

             def a_plus_b(a_from_df: pd.Series, b_from_df: pd.Series) -> pd.Series:
                 return a + b


             # the with_columns call
             @with_columns(
                 load_from=[my_module], # Load from any module
                 columns_to_pass=["a_from_df", "b_from_df"], # The columns to pass from the dataframe to
                 # the subdag
                 select=["a", "b", "a_plus_b"], # The columns to select from the dataframe
             )
             def final_df(initial_df: ps.DataFrame) -> ps.DataFrame:
                 # process, or just return unprocessed
                 ...

         You can think of the above as a series of withColumn calls on the dataframe, where the
         operations are applied in topological order. This is significantly more efficient than
         extracting out the columns, applying the maps, then joining, but *also* allows you to
         express the operations individually, making it easy to unit-test and reuse.

         Note that the operation is "append", meaning that the columns that are selected are appended
         onto the dataframe. We will likely add an option to have this be either "select" or "append"
         mode.

         If the function takes multiple dataframes, the dataframe input to process will always be
         the first one. This will be passed to the subdag, transformed, and passed back to the functions.
         This follows the hamilton rule of reference by parameter name. To demonstarte this, in the code
         above, the dataframe that is passed to the subdag is `initial_df`. That is transformed
         by the subdag, and then returned as the final dataframe.

         You can read it as:

         "final_df is a function that transforms the upstream dataframe initial_df, running the transformations
         from my_module. It starts with the columns a_from_df and b_from_df, and then adds the columns
         a, b, and a_plus_b to the dataframe. It then returns the dataframe, and does some processing on it."


        :param load_from: The functions that will be used to generate the group of map operations.
        :param columns_to_pass: The initial schema of the dataframe. This is used to determine which
            upstream inputs should be taken from the dataframe, and which shouldn't. Note that, if this is
            left empty (and external_inputs is as well), we will assume that all dependencies come
            from the dataframe. This cannot be used in conjunction with pass_dataframe_as.
        :param pass_dataframe_as: The name of the dataframe that we're modifying, as known to the subdag.
            If you pass this in, you are responsible for extracting columns out. If not provided, you have
            to pass columns_to_pass in, and we will extract the columns out for you.
        :param select: Outputs to select from the subdag, i.e. functions/module passed int. If this is left
            blank it will add all possible columns from the subdag to the dataframe.
        :param namespace: The namespace of the nodes, so they don't clash with the global namespace
            and so this can be reused. If its left out, there will be no namespace (in which case you'll want
            to be careful about repeating it/reusing the nodes in other parts of the DAG.)
        :param mode: The mode of the operation. This can be either "append" or "select".
            If it is "append", it will keep all original columns in the dataframe, and append what's in select.
            If it is "select", it will do a global select of columns in the dataframe from the `select` parameter.
            Note that, if the `select` parameter is left blank, it will add all columns in the dataframe
            that are in the subdag. This defaults to `append`. If you're using select, use the `@select` decorator
            instead.
        :param config_required: the list of config keys that are required to resolve any functions. Pass in None\
            if you want the functions/modules to have access to all possible config.
        """
        self.subdag_functions = subdag.collect_functions(load_from)
        self.select = select
        self.initial_schema = columns_to_pass
        if (pass_dataframe_as is not None and columns_to_pass is not None) or (
            pass_dataframe_as is None and columns_to_pass is None
        ):
            raise ValueError(
                "You must specify only one of columns_to_pass and "
                "pass_dataframe_as. "
                "This is because specifying pass_dataframe_as injects into "
                "the set of columns, allowing you to perform your own extraction"
                "from the dataframe. We then execute all columns in the sbudag"
                "in order, passing in that initial dataframe. If you want"
                "to reference columns in your code, you'll have to specify "
                "the set of initial columns, and allow the subdag decorator "
                "to inject the dataframe through. The initial columns tell "
                "us which parameters to take from that dataframe, so we can"
                "feed the right data into the right columns."
            )
        self.dataframe_subdag_param = pass_dataframe_as
        self.namespace = namespace
        self.upstream_dependency = dataframe
        self.mode = mode
        self.config_required = config_required

    @staticmethod
    def _prep_nodes(initial_nodes: List[node.Node]) -> List[node.Node]:
        """Prepares nodes by decorating "default" UDFs with transform.
        This allows us to use the sparkify_node function in transforms
        for both the default ones and the decorated ones.

        :param initial_nodes: Initial nodes to prepare
        :return: Prepared nodes
        """
        out = []
        for node_ in initial_nodes:
            if require_columns.is_default_pyspark_udf(node_):
                col = derive_dataframe_parameter_from_node(node_)
                # todo -- wire through config/function correctly
                # the col is the only dataframe paameter so it is the target node
                (node_,) = require_columns(col).transform_node(node_, {}, node_.callable)
            out.append(node_)
        return out

    @staticmethod
    def create_selector_node(
        upstream_name: str, columns: List[str], node_name: str = "select"
    ) -> node.Node:
        """Creates a selector node. The sole job of this is to select just the specified columns.
        Note this is a utility function that's only called here.

        :param upstream_name: Name of the upstream dataframe node
        :param columns: Columns to select
        :param node_name: Name of the node to create
        :return:
        """

        def new_callable(**kwargs) -> DataFrame:
            return kwargs[upstream_name].select(*columns)

        return node.Node(
            name=node_name,
            typ=DataFrame,
            callabl=new_callable,
            input_types={upstream_name: DataFrame},
        )

    @staticmethod
    def create_drop_node(
        upstream_name: str, columns: List[str], node_name: str = "select"
    ) -> node.Node:
        """Creates a drop node. The sole job of this is to drop just the specified columns.
        Note this is a utility function that's only called here.

        :param upstream_name: Name of the upstream dataframe node
        :param columns: Columns to drop
        :param node_name: Name of the node to create
        :return:
        """

        def new_callable(**kwargs) -> DataFrame:
            return kwargs[upstream_name].drop(*columns)

        return node.Node(
            name=node_name,
            typ=DataFrame,
            callabl=new_callable,
            input_types={upstream_name: DataFrame},
        )

    def _validate_dataframe_subdag_parameter(self, nodes: List[node.Node], fn_name: str):
        all_upstream_dataframe_nodes = _identify_upstream_dataframe_nodes(nodes)
        initial_schema = set(self.initial_schema) if self.initial_schema is not None else set()
        candidates_for_upstream_dataframe = set(all_upstream_dataframe_nodes) - set(initial_schema)
        if (
            len(candidates_for_upstream_dataframe) > 1
            or self.dataframe_subdag_param is None
            and len(candidates_for_upstream_dataframe) > 0
        ):
            raise ValueError(
                f"We found multiple upstream dataframe parameters for function: {fn_name} decorated with "
                f"@with_columns. You specified pass_dataframe_as={self.dataframe_subdag_param} as the upstream "
                f"dataframe parameter, which means that your subdag must have exactly {0 if self.dataframe_subdag_param is None else 1} "
                f"upstream dataframe parameters. Instead, we found the following upstream dataframe parameters: {candidates_for_upstream_dataframe}"
            )
        if self.dataframe_subdag_param is not None:
            if len(candidates_for_upstream_dataframe) == 0:
                raise ValueError(
                    f"You specified your set of UDFs to use upstream dataframe parameter: {self.dataframe_subdag_param} "
                    f"for function: {fn_name} decorated with `with_columns`, but we could not find "
                    "that parameter as a dependency of any of the nodes. Note that that dependency "
                    "must be a pyspark dataframe. If you wish, instead, to supply an initial set of "
                    "columns for the upstream dataframe and refer to those columns directly within "
                    "your UDFs, please use columns_to_pass instead of pass_dataframe_as."
                )
            (upstream_dependency,) = list(candidates_for_upstream_dataframe)
            if upstream_dependency != self.dataframe_subdag_param:
                raise ValueError(
                    f"You specified your set of UDFs to use upstream dataframe parameter: {self.dataframe_subdag_param} "
                    f"for function: {fn_name} decorated with `with_columns`, but we found that parameter "
                    f"as a dependency of a node, but it was not the same as the parameter you specified. "
                    f"Instead, we found: {upstream_dependency}."
                )

    def required_config(self) -> List[str]:
        return self.config_required

    def generate_nodes(self, fn: Callable, config: Dict[str, Any]) -> List[node.Node]:
        """Generates nodes in the with_columns groups. This does the following:

        1. Collects all the nodes from the subdag functions
        2. Prunes them to only include the ones that are upstream from the select columns
        3. Sorts them topologically
        4. Creates a new node for each one, injecting the dataframe parameter into the first one
        5. Creates a new node for the final one, injecting the last node into that one
        6. Returns the list of nodes

        :param fn: Function to generate from
        :param config: Config to use for generating/collecting nodes
        :return: List of nodes that this function produces
        """
        namespace = fn.__name__ if self.namespace is None else self.namespace
        initial_nodes = subdag.collect_nodes(config, self.subdag_functions)
        transformed_nodes = with_columns._prep_nodes(initial_nodes)
        self._validate_dataframe_subdag_parameter(transformed_nodes, fn.__qualname__)
        pruned_nodes = prune_nodes(transformed_nodes, self.select)
        if len(pruned_nodes) == 0:
            raise ValueError(
                f"No nodes found upstream from select columns: {self.select} for function: "
                f"{fn.__qualname__}"
            )
        sorted_initial_nodes = graph_functions.topologically_sort_nodes(pruned_nodes)
        output_nodes = []
        inject_parameter = _derive_first_dataframe_parameter_from_fn(fn)
        current_dataframe_node = inject_parameter
        # Columns that it is dependent on could be from the group of transforms created
        columns_produced_within_mapgroup = {node_.name for node_ in pruned_nodes}
        columns_passed_in_from_dataframe = (
            set(self.initial_schema) if self.initial_schema is not None else []
        )
        drop_list = []
        # Or from the dataframe passed in...
        for node_ in sorted_initial_nodes:
            # dependent columns are broken into two sets:
            # 1. Those that come from the group of transforms
            dependent_columns_in_mapgroup = {
                column for column in node_.input_types if column in columns_produced_within_mapgroup
            }
            # 2. Those that come from the dataframe
            dependent_columns_in_dataframe = {
                column for column in node_.input_types if column in columns_passed_in_from_dataframe
            }
            # In the case that we are using pyspark UDFs
            if require_columns.is_decorated_pyspark_udf(node_):
                sparkified = require_columns.sparkify_node(
                    node_,
                    current_dataframe_node,
                    inject_parameter,
                    self.dataframe_subdag_param,
                    dependent_columns_in_mapgroup,
                    dependent_columns_in_dataframe,
                )
            # otherwise we're using pandas/primitive UDFs
            else:
                sparkified = sparkify_node_with_udf(
                    node_,
                    current_dataframe_node,
                    inject_parameter,
                    self.dataframe_subdag_param,
                    dependent_columns_in_mapgroup,
                    dependent_columns_in_dataframe,
                )
            if self.select is not None and sparkified.name not in self.select:
                # we need to create a drop list because we don't want to drop
                # original columns from the DF by accident.
                drop_list.append(sparkified.name)
            output_nodes.append(sparkified)
            current_dataframe_node = sparkified.name
        # We get the final node, which is the function we're using
        # and reassign inputs to be the dataframe
        if self.mode == "select":
            # this selects over the original DF and the additions
            select_columns = (
                self.select if self.select is not None else [item.name for item in output_nodes]
            )
            select_node = with_columns.create_selector_node(
                upstream_name=current_dataframe_node,
                columns=select_columns,
                node_name="_select",
            )
            output_nodes.append(select_node)
            current_dataframe_node = select_node.name
        elif self.select is not None and len(drop_list) > 0:
            # since it's in append mode, we only want to append what's in the select
            # but we don't know what the original schema is, so we instead drop
            # things from the DF to achieve the same result
            select_node = with_columns.create_drop_node(
                upstream_name=current_dataframe_node,
                columns=drop_list,
                node_name="_select",
            )
            output_nodes.append(select_node)
            current_dataframe_node = select_node.name
        output_nodes = subdag.add_namespace(output_nodes, namespace)
        final_node = node.Node.from_fn(fn).reassign_inputs(
            {inject_parameter: assign_namespace(current_dataframe_node, namespace)}
        )
        return output_nodes + [final_node]

    def validate(self, fn: Callable):
        _derive_first_dataframe_parameter_from_fn(fn)


class select(with_columns):
    def __init__(
        self,
        *load_from: Union[Callable, ModuleType],
        columns_to_pass: List[str] = None,
        pass_dataframe_as: str = None,
        output_cols: List[str] = None,
        namespace: str = None,
        config_required: List[str] = None,
    ):
        """Initializes a select decorator for spark. This allows you to efficiently run
         groups of map operations on a dataframe, represented as pandas/primitives UDFs. This
         effectively "linearizes" compute -- meaning that a DAG of map operations can be run
         as a set of `.select` operations on a single dataframe -- ensuring that you don't have
         to do a complex `extract` then `join` process on spark, which can be inefficient.

         Here's an example of calling it -- if you've seen `@subdag`, you should be familiar with
         the concepts:

         .. code-block:: python

             # my_module.py
             def a(a_from_df: pd.Series) -> pd.Series:
                 return _process(a)

             def b(b_from_df: pd.Series) -> pd.Series:
                 return _process(b)

             def a_plus_b(a_from_df: pd.Series, b_from_df: pd.Series) -> pd.Series:
                 return a + b


             # the with_columns call
             @select(
                 load_from=[my_module], # Load from any module
                 columns_to_pass=["a_from_df", "b_from_df"], # The columns to pass from original dataframe to
                 output_cols=["a", "b", "a_plus_b"], # The columns to have in the final dataframe
             )
             def final_df(initial_df: ps.DataFrame) -> ps.DataFrame:
                 # process, or just return unprocessed
                 ...

         You can think of the above as a series of select/withColumns calls on the dataframe, where the
         operations are applied in topological order. This is significantly more efficient than
         extracting out the columns, applying the maps, then joining, but *also* allows you to
         express the operations individually, making it easy to unit-test and reuse.

         Note that the operation is "append", meaning that the columns that are selected are appended
         onto the dataframe, and then at the end only what is request is selected.

         If the function takes multiple dataframes, the dataframe input to process will always be
         the first one. This will be passed to the subdag, transformed, and passed back to the functions.
         This follows the hamilton rule of reference by parameter name. To demonstarte this, in the code
         above, the dataframe that is passed to the subdag is `initial_df`. That is transformed
         by the subdag, and then returned as the final dataframe.

         You can read it as:

         "final_df is a function that transforms the upstream dataframe initial_df, running the transformations
         from my_module. It starts with the columns a_from_df and b_from_df, and then adds the columns
         a, b, and a_plus_b to the dataframe. It then returns the dataframe, and does some processing on it."


        :param load_from: The functions that will be used to generate the group of map operations.
        :param columns_to_pass: The initial schema of the dataframe. This is used to determine which
            upstream inputs should be taken from the dataframe, and which shouldn't. Note that, if this is
            left empty (and external_inputs is as well), we will assume that all dependencies come
            from the dataframe. This cannot be used in conjunction with pass_dataframe_as.
        :param pass_dataframe_as: The name of the dataframe that we're modifying, as known to the subdag.
            If you pass this in, you are responsible for extracting columns out. If not provided, you have
            to pass columns_to_pass in, and we will extract the columns out for you.
        :param output_cols: Columns to select in the final dataframe. If this is left blank it will
            add all possible columns from the subdag to the dataframe.
        :param namespace: The namespace of the nodes, so they don't clash with the global namespace
            and so this can be reused. If its left out, there will be no namespace (in which case you'll want
            to be careful about repeating it/reusing the nodes in other parts of the DAG.)
        :param config_required: the list of config keys that are required to resolve any functions. Pass in None\
            if you want the functions/modules to have access to all possible config.
        """
        super(select, self).__init__(
            *load_from,
            columns_to_pass=columns_to_pass,
            pass_dataframe_as=pass_dataframe_as,
            select=output_cols,
            namespace=namespace,
            mode="select",
            config_required=config_required,
        )


class SparkInputValidator(lifecycle_base.BaseDoValidateInput):
    """This is a graph hook adapter that allows you to get past a <4.0.0 limitation in spark.
    Spark has the option to choose between spark connect and spark, which largely have the same API.
    That said, they don't have the proper subclass relationships, which make hamilton fail on the input type checking.

    See the following for more information as to why this is necessary:
    - https://community.databricks.com/t5/data-engineering/pyspark-sql-connect-dataframe-dataframe-vs-pyspark-sql-dataframe/td-p/71055
    - https://issues.apache.org/jira/browse/SPARK-47909

    You can access an instance of this through the convenience variable `SPARK_INPUT_CHECK`.
    This allows you to bypass that. This has to be used with the driver builder pattern -- this will look as follows:

    .. code-block:: python

        from hamilton import driver
        from hamilton.plugins import h_spark

        dr = driver.Builder().with_modules(...).with_adapters(h_spark.SPARK_INPUT_CHECK).build()

    Then run it as you would normally. Note that in spark==4.0.0, you will only need the spark session check,
    not the dataframe check.
    """

    def do_validate_input(self, *, node_type: type, input_value: Any) -> bool:
        """Validates the input. Treats connect/classic sessios/dataframe as interchangeable."""

        try:
            from pyspark.sql.connect.dataframe import DataFrame as CDataFrame
            from pyspark.sql.connect.session import SparkSession as CSparkSession
        except ImportError as e:
            raise ImportError(
                "You need to have the spark-connect package installed to use this adapter. That said,"
                "you really shouldn't need this adapter if you don't have it installed (as you won't be able to "
                "get a spark connect session/dataframe without it...). Probably best to remove this"
                "adapter all together, but if you need to you can run pip install 'pyspark[connect]'"
            ) from e
        node_type_is_df = isinstance(input_value, (CDataFrame, DataFrame))
        if node_type_is_df and issubclass(node_type, (DataFrame, CDataFrame)):
            return True
        node_type_is_spark_session = isinstance(input_value, (CSparkSession, SparkSession))
        if node_type_is_spark_session and issubclass(node_type, (CSparkSession, SparkSession)):
            return True
        return htypes.check_input_type(node_type, input_value)


SPARK_INPUT_CHECK = SparkInputValidator()



import asyncio
import inspect
import logging
import sys
import time
import typing
import uuid
from types import ModuleType
from typing import Any, Dict, Optional, Tuple

import hamilton.lifecycle.base as lifecycle_base
from hamilton import base, driver, graph, lifecycle, node, telemetry
from hamilton.execution.graph_functions import create_error_message
from hamilton.io.materialization import ExtractorFactory, MaterializerFactory

logger = logging.getLogger(__name__)


async def await_dict_of_tasks(task_dict: Dict[str, typing.Awaitable]) -> Dict[str, Any]:
    """Util to await a dictionary of tasks as asyncio.gather is kind of garbage"""
    keys = sorted(task_dict.keys())
    coroutines = [task_dict[key] for key in keys]
    coroutines_gathered = await asyncio.gather(*coroutines)
    return dict(zip(keys, coroutines_gathered))


async def process_value(val: Any) -> Any:
    """Helper function to process the value of a potential awaitable.
    This is very simple -- all it does is await the value if its not already resolved.

    :param val: Value to process.
    :return: The value (awaited if it is a coroutine, raw otherwise).
    """
    if not inspect.isawaitable(val):
        return val
    return await val


class AsyncGraphAdapter(lifecycle_base.BaseDoNodeExecute, lifecycle.ResultBuilder):
    """Graph adapter for use with the :class:`AsyncDriver` class."""

    def __init__(
        self,
        result_builder: base.ResultMixin = None,
        async_lifecycle_adapters: Optional[lifecycle_base.LifecycleAdapterSet] = None,
    ):
        """Creates an AsyncGraphAdapter class. Note this will *only* work with the AsyncDriver class.

        Some things to note:

            1. This executes everything at the end (recursively). E.G. the final DAG nodes are awaited
            2. This does *not* work with decorators when the async function is being decorated. That is\
            because that function is called directly within the decorator, so we cannot await it.
        """
        super(AsyncGraphAdapter, self).__init__()
        self.adapter = (
            async_lifecycle_adapters
            if async_lifecycle_adapters is not None
            else lifecycle_base.LifecycleAdapterSet()
        )
        self.result_builder = result_builder if result_builder else base.PandasDataFrameResult()
        self.is_initialized = False

    def do_node_execute(
        self,
        *,
        run_id: str,
        node_: node.Node,
        kwargs: typing.Dict[str, typing.Any],
        task_id: Optional[str] = None,
    ) -> typing.Any:
        """Executes a node. Note this doesn't actually execute it -- rather, it returns a task.
        This does *not* use async def, as we want it to be awaited on later -- this await is done
        in processing parameters of downstream functions/final results. We can ensure that as
        we also run the driver that this corresponds to.

        Note that this assumes that everything is awaitable, even if it isn't.
        In that case, it just wraps it in one.

        :param task_id:
        :param node_:
        :param run_id:
        :param node: Node to wrap
        :param kwargs: Keyword arguments (either coroutines or raw values) to call it with
        :return: A task
        """
        callabl = node_.callable

        async def new_fn(fn=callabl, **fn_kwargs):
            task_dict = {key: process_value(value) for key, value in fn_kwargs.items()}
            fn_kwargs = await await_dict_of_tasks(task_dict)
            error = None
            result = None
            success = True
            pre_node_execute_errored = False
            try:
                if self.adapter.does_hook("pre_node_execute", is_async=True):
                    try:
                        await self.adapter.call_all_lifecycle_hooks_async(
                            "pre_node_execute",
                            run_id=run_id,
                            node_=node_,
                            kwargs=fn_kwargs,
                            task_id=task_id,
                        )
                    except Exception as e:
                        pre_node_execute_errored = True
                        raise e
                # TODO -- consider how to use node execution methods in the future
                # This is messy as it is a method called within a method...
                # if self.adapter.does_method("do_node_execute", is_async=False):
                #     result = self.adapter.call_lifecycle_method_sync(
                #         "do_node_execute",
                #         run_id=run_id,
                #         node_=node_,
                #         kwargs=kwargs,
                #         task_id=task_id,
                #     )
                # else:

                result = (
                    await fn(**fn_kwargs) if asyncio.iscoroutinefunction(fn) else fn(**fn_kwargs)
                )
            except Exception as e:
                success = False
                error = e
                step = "[pre-node-execute:async]" if pre_node_execute_errored else ""
                message = create_error_message(kwargs, node_, step)
                logger.exception(message)
                raise
            finally:
                if not pre_node_execute_errored and self.adapter.does_hook(
                    "post_node_execute", is_async=True
                ):
                    try:
                        await self.adapter.call_all_lifecycle_hooks_async(
                            "post_node_execute",
                            run_id=run_id,
                            node_=node_,
                            kwargs=fn_kwargs,
                            success=success,
                            error=error,
                            result=result,
                            task_id=task_id,
                        )
                    except Exception:
                        message = create_error_message(kwargs, node_, "[post-node-execute]")
                        logger.exception(message)
                        raise

            return result

        coroutine = new_fn(**kwargs)
        task = asyncio.create_task(coroutine)
        return task

    def build_result(self, **outputs: Any) -> Any:
        return self.result_builder.build_result(**outputs)


def separate_sync_from_async(
    adapters: typing.List[lifecycle.LifecycleAdapter],
) -> Tuple[typing.List[lifecycle.LifecycleAdapter], typing.List[lifecycle.LifecycleAdapter]]:
    """Separates the sync and async adapters from a list of adapters.
    Note this only works with hooks -- we'll be dealing with methods later.

    :param adapters: List of adapters
    :return: Tuple of sync adapters, async adapters
    """

    adapter_set = lifecycle_base.LifecycleAdapterSet(*adapters)
    # this is using internal(ish) fields (.sync_hooks/.async_hooks) -- we should probably expose it
    # For now this is OK
    # Note those are dict[hook_name, list[hook]], so we have to flatten
    return (
        [sync_adapter for adapters in adapter_set.sync_hooks.values() for sync_adapter in adapters],
        [
            async_adapter
            for adapters in adapter_set.async_hooks.values()
            for async_adapter in adapters
        ],
    )


class AsyncDriver(driver.Driver):
    """Async driver. This is a driver that uses the AsyncGraphAdapter to execute the graph.

    .. code-block:: python

        dr = async_driver.AsyncDriver({}, async_module, result_builder=base.DictResult())
        df = await dr.execute([...], inputs=...)

    """

    def __init__(
        self,
        config,
        *modules,
        result_builder: Optional[base.ResultMixin] = None,
        adapters: typing.List[lifecycle.LifecycleAdapter] = None,
    ):
        """Instantiates an asynchronous driver.

        You will also need to call `ainit` to initialize the driver if you have any hooks/adapters.

        Note that this is not the desired API -- you should be using the :py:class:`hamilton.async_driver.Builder` class to create the driver.

        This will only (currently) work properly with asynchronous lifecycle hooks, and does not support methods or validators.
        You can still pass in synchronous lifecycle hooks, but they may behave strangely.

        :param config: Config to build the graph
        :param modules: Modules to crawl for fns/graph nodes
        :param adapters: Adapters to use for lifecycle methods.
        :param result_builder: Results mixin to compile the graph's final results. TBD whether this should be included in the long run.
        """
        if adapters is None:
            adapters = []
        sync_adapters, async_adapters = separate_sync_from_async(adapters)

        # we'll need to use this in multiple contexts so we'll keep it around for later

        result_builders = [adapter for adapter in adapters if isinstance(adapter, base.ResultMixin)]
        if result_builder is not None:
            result_builders.append(result_builder)
        if len(result_builders) > 1:
            raise ValueError(
                "You cannot pass more than one result builder to the async driver. "
                "Please pass in a single result builder"
            )
        # it will be defaulted by the graph adapter
        result_builder = result_builders[0] if len(result_builders) == 1 else None
        super(AsyncDriver, self).__init__(
            config,
            *modules,
            adapter=[
                # We pass in the async adapters here as this can call node-level hooks
                # Otherwise we trust the driver/fn graph to call sync adapters
                AsyncGraphAdapter(
                    result_builder=result_builder,
                    async_lifecycle_adapters=lifecycle_base.LifecycleAdapterSet(*async_adapters),
                ),
                # We pass in the sync adapters here as this can call
                *sync_adapters,
                *async_adapters,  # note async adapters will not be called during synchronous execution -- this is for access later
            ],
        )
        self.initialized = False

    async def ainit(self) -> "AsyncDriver":
        """Initializes the driver when using async. This only exists for backwards compatibility.
        In Hamilton 2.0, we will be using an asynchronous constructor.
        See https://dev.to/akarshan/asynchronous-python-magic-how-to-create-awaitable-constructors-with-asyncmixin-18j5.
        """
        if self.initialized:
            # this way it can be called twice
            return self
        if self.adapter.does_hook("post_graph_construct", is_async=True):
            await self.adapter.call_all_lifecycle_hooks_async(
                "post_graph_construct",
                graph=self.graph,
                modules=self.graph_modules,
                config=self.config,
            )
        await self.adapter.ainit()
        self.initialized = True
        return self

    async def raw_execute(
        self,
        final_vars: typing.List[str],
        overrides: Dict[str, Any] = None,
        display_graph: bool = False,  # don't care
        inputs: Dict[str, Any] = None,
        _fn_graph: graph.FunctionGraph = None,
    ) -> Dict[str, Any]:
        """Executes the graph, returning a dictionary of strings (node keys) to final results.

        :param final_vars: Variables to execute (+ upstream)
        :param overrides: Overrides for nodes
        :param display_graph: whether or not to display graph -- this is not supported.
        :param inputs:  Inputs for DAG runtime calculation
        :param _fn_graph: Function graph for compatibility with superclass -- unused
        :return: A dict of key -> result
        """
        assert _fn_graph is None, (
            "_fn_graph must not be provided "
            "-- the only reason you'd do this is to use materialize(), which is not supported yet.."
        )
        run_id = str(uuid.uuid4())
        nodes, user_nodes = self.graph.get_upstream_nodes(final_vars, inputs, overrides)
        memoized_computation = dict()  # memoized storage
        if self.adapter.does_hook("pre_graph_execute"):
            await self.adapter.call_all_lifecycle_hooks_sync_and_async(
                "pre_graph_execute",
                run_id=run_id,
                graph=self.graph,
                final_vars=final_vars,
                inputs=inputs,
                overrides=overrides,
            )
        results = None
        error = None
        success = False
        try:
            self.graph.execute(nodes, memoized_computation, overrides, inputs, run_id=run_id)
            if display_graph:
                raise ValueError(
                    "display_graph=True is not supported for the async graph adapter. "
                    "Instead you should be using visualize_execution."
                )
            task_dict = {
                key: asyncio.create_task(process_value(memoized_computation[key]))
                for key in final_vars
            }
            results = await await_dict_of_tasks(task_dict)
            success = True
        except Exception as e:
            error = e
            success = False
            raise e
        finally:
            if self.adapter.does_hook("post_graph_execute", is_async=None):
                await self.adapter.call_all_lifecycle_hooks_sync_and_async(
                    "post_graph_execute",
                    run_id=run_id,
                    graph=self.graph,
                    success=success,
                    error=error,
                    results=results,
                )
        return results

    async def execute(
        self,
        final_vars: typing.List[str],
        overrides: Dict[str, Any] = None,
        display_graph: bool = False,
        inputs: Dict[str, Any] = None,
    ) -> Any:
        """Executes computation.

        :param final_vars: the final list of variables we want to compute.
        :param overrides: values that will override "nodes" in the DAG.
        :param display_graph: DEPRECATED. Whether we want to display the graph being computed.
        :param inputs: Runtime inputs to the DAG.
        :return: an object consisting of the variables requested, matching the type returned by the GraphAdapter.
            See constructor for how the GraphAdapter is initialized. The default one right now returns a pandas
            dataframe.
        """
        if display_graph:
            raise ValueError(
                "display_graph=True is not supported for the async graph adapter. "
                "Instead you should be using visualize_execution."
            )
        start_time = time.time()
        run_successful = True
        error = None
        _final_vars = self._create_final_vars(final_vars)
        try:
            outputs = await self.raw_execute(_final_vars, overrides, display_graph, inputs=inputs)
            # Currently we don't allow async build results, but we could.
            if self.adapter.does_method("do_build_result", is_async=False):
                return self.adapter.call_lifecycle_method_sync("do_build_result", outputs=outputs)
            return outputs
        except Exception as e:
            run_successful = False
            logger.error(driver.SLACK_ERROR_MESSAGE)
            error = telemetry.sanitize_error(*sys.exc_info())
            raise e
        finally:
            duration = time.time() - start_time
            # ensure we can capture telemetry in async friendly way.
            if telemetry.is_telemetry_enabled():

                async def make_coroutine():
                    self.capture_execute_telemetry(
                        error, final_vars, inputs, overrides, run_successful, duration
                    )

                try:
                    # we don't have to await because we are running within the event loop.
                    asyncio.create_task(make_coroutine())
                except Exception as e:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.error(f"Encountered error submitting async telemetry:\n{e}")

    def capture_constructor_telemetry(
        self,
        error: Optional[str],
        modules: Tuple[ModuleType],
        config: Dict[str, Any],
        adapter: base.HamiltonGraphAdapter,
    ):
        """Ensures we capture constructor telemetry the right way in an async context.

        This is a simpler wrapper around what's in the driver class.

        :param error: sanitized error string, if any.
        :param modules: tuple of modules to build DAG from.
        :param config: config to create the driver.
        :param adapter: adapter class object.
        """
        if telemetry.is_telemetry_enabled():
            try:
                # check whether the event loop has been started yet or not
                loop = asyncio.get_event_loop()
                if loop.is_running():
                    loop.run_in_executor(
                        None,
                        super(AsyncDriver, self).capture_constructor_telemetry,
                        error,
                        modules,
                        config,
                        adapter,
                    )
                else:

                    async def make_coroutine():
                        super(AsyncDriver, self).capture_constructor_telemetry(
                            error, modules, config, adapter
                        )

                    loop.run_until_complete(make_coroutine())
            except Exception as e:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.error(f"Encountered error submitting async telemetry:\n{e}")


class Builder(driver.Builder):
    """Builder for the async driver. This is equivalent to the standard builder, but has a more limited API.
    Note this does not support dynamic execution or materializers (for now).

    Here is an example of how you might use it to get the tracker working:

    .. code-block:: python

        from hamilton_sdk import tracker

        tracker_async = adapters.AsyncHamiltonTracker(
            project_id=1,
            username="elijah",
            dag_name="async_tracker",
        )
        dr = (
            await async_driver.Builder()
            .with_modules(async_module)
            .with_adapters(tracking_async)
            .build()
        )
    """

    def __init__(self):
        super(Builder, self).__init__()

    def _not_supported(self, method_name: str, additional_message: str = ""):
        raise ValueError(
            f"Builder().{method_name}() is not supported for the async driver. {additional_message}"
        )

    def enable_dynamic_execution(self, *, allow_experimental_mode: bool = False) -> "Builder":
        self._not_supported("enable_dynamic_execution")

    def with_materializers(
        self, *materializers: typing.Union[ExtractorFactory, MaterializerFactory]
    ) -> "Builder":
        self._not_supported("with_materializers")

    def with_adapter(self, adapter: base.HamiltonGraphAdapter) -> "Builder":
        self._not_supported(
            "with_adapter",
            "Use with_adapters instead to pass in the tracker (or other async hooks/methods)",
        )

    def build_without_init(self) -> AsyncDriver:
        """Allows you to build the async driver without initialization. Use this at
        your own risk -- we highly recommend calling `.ainit` on the final result.

        :return:
        """
        adapters = self.adapters if self.adapters is not None else []
        if self.legacy_graph_adapter is not None:
            adapters.append(self.legacy_graph_adapter)

        # We should really be doing this in the constructor
        # but the AsyncGraphAdapter originally used the pandas builder
        # so we pass in the right one to ensure backwards compatibility
        # This will become the default API soon, so it's OK to put the complexity here
        result_builders = [adapter for adapter in adapters if isinstance(adapter, base.ResultMixin)]
        specified_result_builder = base.DictResult() if len(result_builders) == 0 else None
        return AsyncDriver(
            self.config,
            *self.modules,
            adapters=self.adapters,
            result_builder=specified_result_builder,
        )

    async def build(self):
        """Builds the async driver. This also initializes it, hence the async definition.
        If you don't want to use async, you can use `build_without_init` and call `ainit` later,
        but we recommend using this in an asynchronous lifespan management function (E.G. in fastAPI),
        or something similar.

        :return: The fully
        """
        dr = self.build_without_init()
        return await dr.ainit()



import enum
import functools
import json
import logging
from pathlib import Path
from typing import Any, Dict, Literal, Mapping, NamedTuple, Union

import pyarrow
import pyarrow.ipc
from pyarrow.interchange import from_dataframe

from hamilton.experimental import h_databackends
from hamilton.graph_types import HamiltonGraph, HamiltonNode
from hamilton.lifecycle import GraphExecutionHook, NodeExecutionHook

logger = logging.getLogger(__name__)


class Diff(enum.Enum):
    """Result of a diff operation: ADDED, REMOVED, EQUAL, or UNEQUAL. They are mutually exclusive."""

    ADDED = "+"
    REMOVED = "-"
    EQUAL = "=="
    UNEQUAL = "!="


class DiffResult(NamedTuple):
    diff: Diff
    value: Any


def _diff_mappings(
    current: Mapping[str, Any], reference: Mapping[str, Any]
) -> Dict[str, DiffResult]:
    """Generate the diff for all fields of two mappings.

    example:
        {
            "foo": {DiffResult.ADDED: "foo_value"},
            "bar": {DiffResult.REMOVED: "bar_value"},
            "baz": {DiffResult.EQUAL: "baz_value"},
            "oof": {DiffResult.UNEQUAL: {"cur": "oof_value", "ref": "old_value"}}
        }
    """
    current_key_set = set(current.keys())
    reference_key_set = set(reference.keys())

    current_only = current_key_set.difference(reference_key_set)
    reference_only = reference_key_set.difference(current_key_set)
    shared = current_key_set.intersection(reference_key_set)

    diff = {}
    for key in current_only:
        diff[key] = DiffResult(Diff.ADDED, current[key])

    for key in reference_only:
        diff[key] = DiffResult(Diff.REMOVED, reference[key])

    for key in shared:
        current_value = current[key]
        reference_value = reference[key]

        if current_value == reference_value:
            diff[key] = DiffResult(Diff.EQUAL, current_value)
        else:
            diff[key] = DiffResult(Diff.UNEQUAL, {"cur": current_value, "ref": reference_value})

    return diff


SCHEMA_METADATA_FIELD = "__metadata"


# TODO add check for field metadata
def diff_schemas(
    current_schema: pyarrow.Schema,
    reference_schema: pyarrow.Schema,
    check_schema_metadata: bool = False,
    check_field_metadata: bool = False,
):
    """Diff two Pyarrow schema field-by-field. Options to diff schema and field metadata key-by-key.
    Returning an empty dict means equality / no diff

    example:
        {
            '__metadata': DiffResult(diff=<Diff.UNEQUAL: '!='>, value={
                'key': DiffResult(diff=<Diff.UNEQUAL: '!='>, value={
                    'cur': 'value1', 'ref': 'value2'
                })
            }),
            'bar': DiffResult(diff=<Diff.EQUAL: '=='>, value={
                'name': 'bar', 'type': 'int64', 'nullable': True, 'metadata': {}
            }),
            'foo': DiffResult(diff=<Diff.UNEQUAL: '!='>, value={
                'name': DiffResult(diff=<Diff.EQUAL: '=='>, value='foo')
                'type': DiffResult(diff=<Diff.EQUAL: '=='>, value='string'),
                'nullable': DiffResult(diff=<Diff.EQUAL: '=='>, value=True),
                'metadata': DiffResult(diff=<Diff.UNEQUAL: '!='>, value={
                    'key': DiffResult(diff=<Diff.UNEQUAL: '!='>, value={
                        'cur': 'value1', 'ref': 'value2'
                    })
                }),
            })
        }

    """
    # if schemas are equal, return an empty diff
    if current_schema.equals(
        reference_schema, check_metadata=(check_schema_metadata or check_field_metadata)
    ):
        return {}

    current_schema = pyarrow_schema_to_json(current_schema)
    reference_schema = pyarrow_schema_to_json(reference_schema)

    schema_diff = _diff_mappings(current=current_schema, reference=reference_schema)

    # compare fields shared by both schemas
    for field_name, diff_result in schema_diff.items():
        if diff_result.diff == Diff.UNEQUAL:
            current_field = current_schema.get(field_name, {})
            reference_field = reference_schema.get(field_name, {})

            field_diff = DiffResult(
                diff=Diff.UNEQUAL,
                value=_diff_mappings(current=current_field, reference=reference_field),
            )

            if check_field_metadata and (field_diff.value["metadata"].diff != Diff.EQUAL):
                current_field_metadata = current_field.get("metadata", {})
                reference_field_metadata = reference_field.get("metadata", {})

                field_diff.value["metadata"] = DiffResult(
                    diff=Diff.UNEQUAL,
                    value=_diff_mappings(
                        current=current_field_metadata,
                        reference=reference_field_metadata,
                    ),
                )

            schema_diff[field_name] = field_diff

    # compare schema metadata
    if check_schema_metadata:
        current_schema_metadata = current_schema.get(SCHEMA_METADATA_FIELD, {})
        reference_schema_metadata = reference_schema.get(SCHEMA_METADATA_FIELD, {})

        schema_diff[SCHEMA_METADATA_FIELD] = DiffResult(
            Diff.UNEQUAL,
            _diff_mappings(current=current_schema_metadata, reference=reference_schema_metadata),
        )

    return schema_diff


def human_readable_diff(diff: dict) -> dict:
    """Format a diff to exclude EQUAL fields and make it easier to read.

    example:
        {
            '__metadata': {
                'key': {'cur': 'value1', 'ref': 'value2'}
            },
            "foo": {
                "metadata": {
                    'key': {"cur": "value1", "ref": "value2"}
                },
            },
            "baz": "-",
            "bar": {
                "type": {"cur": "int64", "ref": "double"}
            },
        }
    """

    readable_diff = {}

    for field_name, diff_result in diff.items():
        # special case for the schema metadata field
        if field_name == SCHEMA_METADATA_FIELD:
            schema_metadata_diff = human_readable_diff(diff_result.value)
            if schema_metadata_diff != {}:
                readable_diff[SCHEMA_METADATA_FIELD] = schema_metadata_diff
            continue

        if diff_result.diff == Diff.EQUAL:
            continue

        elif diff_result.diff == Diff.ADDED:
            readable_diff[field_name] = diff_result.diff.value

        elif diff_result.diff == Diff.REMOVED:
            readable_diff[field_name] = diff_result.diff.value

        elif diff_result.diff == Diff.UNEQUAL:
            if diff_result.value.get("cur"):
                readable_diff[field_name] = diff_result.value
            else:
                readable_diff[field_name] = human_readable_diff(diff_result.value)

    return dict(sorted(readable_diff.items()))


def pyarrow_schema_to_json(schema: pyarrow.Schema) -> dict:
    """Convert a pyarrow.Schema to a JSON-serializable dictionary

    Pyarrow provides a schema-to-string, but not schema-to-json.
    """

    schema_dict = dict()

    if schema.metadata:
        schema_dict[SCHEMA_METADATA_FIELD] = {
            k.decode(): v.decode() for k, v in schema.metadata.items()
        }

    for name in schema.names:
        field = schema.field(name)
        field_metadata = {}
        if field.metadata:
            field_metadata = {k.decode(): v.decode() for k, v in field.metadata.items()}

        schema_dict[str(name)] = dict(
            name=field.name,
            type=field.type.__str__(),  # __str__() and __repr__() of pyarrow.Field are different
            nullable=field.nullable,
            metadata=field_metadata,
        )

    return schema_dict


@functools.singledispatch
def _get_arrow_schema(df, allow_copy: bool = True) -> pyarrow.Schema:
    """Base case for getting a pyarrow schema from a dataframe.

    :param allow_copy: If True, allow to convert the object to Pyarrow
        even if zero-copy is unavailable

    It looks for the `__dataframe__` attribute associated with the dataframe interchange protocol
    ref: https://data-apis.org/dataframe-protocol/latest/API.html
    """
    if not hasattr(df, "__dataframe__"):
        # if hitting this condition, we can register a new function
        # to conver the dataframe to pyarrow
        raise NotImplementedError(f"Type {type(df)} is currently unsupported.")

    # try to convert to Pyarrow using zero-copy
    try:
        df = from_dataframe(df, allow_copy=False)
    # if unable to zero-copy, convert the object to Pyarrow
    # this may be undesirable if the object is large because of memory overhead
    except RuntimeError as e:
        if allow_copy is False:
            raise e
        df = from_dataframe(df, allow_copy=True)
    return df.schema


@_get_arrow_schema.register
def _(df: h_databackends.AbstractPandasDataFrame, **kwargs) -> pyarrow.Schema:
    """pandas to pyarrow using pyarrow-native method.
    Removes the pandas metadata added by Pyarrow for cleaner schema diffs
    """
    table = pyarrow.Table.from_pandas(df, preserve_index=False)
    return table.schema.remove_metadata()


@_get_arrow_schema.register
def _(df: h_databackends.AbstractIbisDataFrame, **kwargs) -> pyarrow.Schema:
    """Convert the Ibis schema to pyarrow Schema. The operation is lazy
    and doesn't require Ibis execution"""
    return df.schema().to_pyarrow()


def _spark_to_arrow(type_):
    import pyspark.sql.types as pt

    _from_pyspark_dtypes = {
        pt.NullType: pyarrow.null(),
        pt.BooleanType: pyarrow.bool_(),
        pt.BinaryType: pyarrow.binary(),
        pt.ByteType: pyarrow.int8(),
        pt.ShortType: pyarrow.int16(),
        pt.IntegerType: pyarrow.int32(),
        pt.LongType: pyarrow.int64(),
        pt.DateType: pyarrow.date64(),
        pt.FloatType: pyarrow.float32(),
        pt.DoubleType: pyarrow.float64(),
        pt.TimestampType: pyarrow.timestamp(unit="ms", tz=None),
        pt.TimestampNTZType: pyarrow.timestamp(unit="ms", tz=None),
        pt.StringType: pyarrow.string(),
        pt.VarcharType: pyarrow.string(),  # TODO specify length
        pt.CharType: pyarrow.string(),  # TODO specify length
        pt.DayTimeIntervalType: pyarrow.month_day_nano_interval(),  # TODO specify unit
        pt.YearMonthIntervalType: pyarrow.month_day_nano_interval(),  # TODO specify unit
    }

    if isinstance(type_, pt.DecimalType):
        arrow_type = pyarrow.decimal128(type_.precision, type_.scale)
    elif isinstance(type_, pt.ArrayType):
        arrow_type = pyarrow.array([], type=_spark_to_arrow(type_.elementType))
    elif isinstance(type_, pt.MapType):
        arrow_type = pyarrow.map_(
            _spark_to_arrow(type_.keyType),
            _spark_to_arrow(type_.valueType),
        )
    elif isinstance(type_, pt.StructType):
        arrow_type = pyarrow.struct(
            {field.name: _spark_to_arrow(field.dataType) for field in type_.fields}
        )
    else:
        try:
            arrow_type = _from_pyspark_dtypes[type(type_)]
        except KeyError:
            raise NotImplementedError(f"Can't convert {type_} to pyarrow type.")

    return arrow_type


@_get_arrow_schema.register(h_databackends.AbstractSparkSQLDataFrame)
def _get_spark_schema(df, **kwargs) -> pyarrow.Schema:
    """Convert the PySpark schema to pyarrow Schema. The operation is lazy
    and doesn't require PySpark execution"""
    return pyarrow.schema(
        pyarrow.field(
            name=field.name,
            type=_spark_to_arrow(field.dataType),
            nullable=field.nullable,
            metadata=field.metadata,
        )
        for field in df.schema
    )


# TODO lazy polars schema conversion
# ongoing polars discussion: https://github.com/pola-rs/polars/issues/15600


def get_dataframe_schema(
    df: Union[h_databackends.DATAFRAME_TYPES], node: HamiltonNode
) -> pyarrow.Schema:
    """Get pyarrow schema of a node result and store node metadata on the pyarrow schema."""
    schema = _get_arrow_schema(df)
    metadata = dict(
        name=str(node.name),
        documentation=str(node.documentation),
        version=str(node.version),
    )
    return schema.with_metadata(metadata)


def load_schema(path: Union[str, Path]) -> pyarrow.Schema:
    """Load pyarrow schema from disk using IPC deserialization"""
    return pyarrow.ipc.read_schema(path)


def save_schema(path: Union[str, Path], schema: pyarrow.Schema) -> None:
    """Save pyarrow schema to disk using IPC serialization"""
    Path(path).write_bytes(schema.serialize())


class SchemaValidator(NodeExecutionHook, GraphExecutionHook):
    """Collect dataframe and columns schemas at runtime. Can also conduct runtime checks against schemas"""

    def __init__(
        self,
        schema_dir: str = "./schemas",
        check: bool = True,
        must_exist: bool = False,
        importance: Literal["warn", "fail"] = "warn",
        check_schema_metadata: bool = False,
        check_field_metadata: bool = False,
    ):
        """
        :param schema_dir: Directory where schema files will be stored.
        :param check: If True, conduct schema validation
            Else generate schema files, potentially overwriting stored schemas.
        :param must_exist: If True when check=True, raise FileNotFoundError for missing schema files
            Else generate the missing schema files.
        :param importance: How to handle unequal schemas when check=True
            "warn": log a warning with the schema diff
            "fail": raise an exception with the schema diff
        """
        self.schemas: Dict[str, pyarrow.Schema] = {}
        self.reference_schemas: Dict[str, pyarrow.Schema] = {}
        self.schema_diffs: dict = {}
        self.schema_dir = schema_dir
        self.check = check
        self.must_exist = must_exist
        self.importance = importance
        self.check_schema_metadata = check_schema_metadata
        self.check_field_metadata = check_field_metadata
        self.h_graph: HamiltonGraph = None

        # create the directory where schemas will be stored
        Path(schema_dir).mkdir(parents=True, exist_ok=True)

    @property
    def json_schemas(self) -> Dict[str, dict]:
        """Return schemas collected during the run"""
        return {
            node_name: pyarrow_schema_to_json(schema) for node_name, schema in self.schemas.items()
        }

    # TODO support nodes returning columns by writing them as single column dataframe
    def get_schema_path(self, node_name: str) -> Path:
        """Generate schema filepath based on node name.

        The `.schema` extension is arbitrary. Another common choice is `.arrow`
        but it wouldn't communicate that the file contains only schema metadata.
        The serialization format is IPC by default (see `.save_schema()`).
        """
        return Path(self.schema_dir, node_name).with_suffix(".schema")

    def run_before_graph_execution(
        self, *, graph: HamiltonGraph, inputs: Dict[str, Any], overrides: Dict[str, Any], **kwargs
    ):
        """Store schemas of inputs and overrides nodes that are tables or columns."""
        self.h_graph = graph
        for node_sets in [inputs, overrides]:
            if node_sets is None:
                continue

            for node_name, node_value in node_sets.items():
                self.run_after_node_execution(node_name=node_name, result=node_value)

    def run_after_node_execution(self, *, node_name: str, result: Any, **kwargs):
        """Store schema of executed node if table or column type."""
        if not isinstance(result, h_databackends.DATAFRAME_TYPES):
            return

        # generate the schema from the HamiltonNode and node value
        node = self.h_graph[node_name]
        schema = get_dataframe_schema(df=result, node=node)
        self.schemas[node_name] = schema

        schema_path = self.get_schema_path(node_name)

        # behavior 1: only save schema
        if self.check is False:
            save_schema(self.get_schema_path(node_name), schema)
            return

        # behavior 2: handle missing reference schema while validating
        if not schema_path.exists():
            if self.must_exist:
                raise FileNotFoundError(
                    f"{schema_path} not found. Set `check=False` or `must_exist=False` to create it."
                )
            else:
                save_schema(self.get_schema_path(node_name), schema)
                return

        # behavior 3: validate current schema with reference schema
        reference_schema = load_schema(self.get_schema_path(node_name))
        schema_diff = diff_schemas(
            current_schema=schema,
            reference_schema=reference_schema,
            check_schema_metadata=self.check_schema_metadata,
            check_field_metadata=self.check_field_metadata,
        )
        if schema_diff != {}:
            self.schema_diffs[node_name] = schema_diff
            readable_diff = json.dumps({node_name: human_readable_diff(schema_diff)}, indent=2)
            if self.importance == "warn":
                logger.warning(f"Schema diff:\n{readable_diff}\n")
            elif self.importance == "fail":
                raise RuntimeError(readable_diff)

    def run_after_graph_execution(self, *args, **kwargs):
        """Store a human-readable JSON of all current schemas."""
        Path(f"{self.schema_dir}/schemas.json").write_text(json.dumps(self.json_schemas))

    def run_before_node_execution(self, *args, **kwargs):
        """Required by subclassing NodeExecutionHook"""



from pathlib import Path
from types import ModuleType
from typing import Dict, List, Union

from hamilton import driver

CONFIG_HEADER = "HAMILTON_CONFIG"
FINAL_VARS_HEADER = "HAMILTON_FINAL_VARS"
INPUTS_HEADER = "HAMILTON_INPUTS"
OVERRIDES_HEADER = "HAMILTON_OVERRIDES"


def get_git_base_directory() -> str:
    """Get the base path of the current git directory"""
    import subprocess

    try:
        result = subprocess.run(
            ["git", "rev-parse", "--show-toplevel"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )

        if result.returncode == 0:
            return result.stdout.strip()
        else:
            print("Error:", result.stderr.strip())
            raise OSError(f"{result.stderr.strip()}")
    except FileNotFoundError:
        raise FileNotFoundError("Git command not found. Please make sure Git is installed.")


def get_git_reference(git_relative_path: Union[str, Path], git_reference: str) -> str:
    """Get the source code from the specified file and git reference"""
    import subprocess

    try:
        result = subprocess.run(
            ["git", "show", f"{git_reference}:{git_relative_path}"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )

        if result.returncode == 0:
            return result.stdout.strip()
        elif result.returncode == 128:
            # TODO catch the following error:
            # fatal: Path '{git_relative_path}' exists on disk, but not in '{git_revision}'.
            return
        else:
            return
    except FileNotFoundError:
        raise FileNotFoundError("Git command not found. Please make sure Git is installed.")


def version_hamilton_functions(module: ModuleType) -> Dict[str, str]:
    """Hash the source code of Hamilton functions from a module"""
    from hamilton import graph_types, graph_utils

    origins_version: Dict[str, str] = dict()

    for origin_name, _ in graph_utils.find_functions(module):
        origin_callable = getattr(module, origin_name)
        origins_version[origin_name] = graph_types.hash_source_code(origin_callable, strip=True)

    return origins_version


def hash_hamilton_nodes(dr: driver.Driver) -> Dict[str, str]:
    """Hash the source code of Hamilton functions from nodes in a Driver"""
    from hamilton import graph_types

    graph = graph_types.HamiltonGraph.from_graph(dr.graph)
    return {n.name: n.version for n in graph.nodes}


def map_nodes_to_functions(dr: driver.Driver) -> Dict[str, str]:
    """Get a mapping from node name to Hamilton function name"""
    from hamilton import graph_types

    graph = graph_types.HamiltonGraph.from_graph(dr.graph)

    node_to_function = dict()
    for n in graph.nodes:
        # is None for config nodes
        if n.originating_functions is None:
            continue

        node_callable = n.originating_functions[0]
        node_to_function[n.name] = node_callable.__name__

    return node_to_function


def hash_dataflow(nodes_version: Dict[str, str]) -> str:
    """Create a dataflow hash from the hashes of its nodes"""
    import hashlib

    sorted_nodes = sorted(nodes_version.values())
    return hashlib.sha256(str(sorted_nodes).encode()).hexdigest()


def load_modules_from_git(
    module_paths: List[Path], git_reference: str = "HEAD"
) -> List[ModuleType]:
    """Dynamically import modules for a git reference"""
    from hamilton import ad_hoc_utils

    git_base_dir = Path(get_git_base_directory())

    modules = []
    for module_path in module_paths:
        relative_to_git = module_path.relative_to(git_base_dir)
        previous_revision = get_git_reference(relative_to_git, git_reference)
        module = ad_hoc_utils.module_from_source(previous_revision)
        modules.append(module)

    return modules


def diff_nodes_against_functions(
    nodes_version: Dict[str, str],
    origins_version: Dict[str, str],
    node_to_origin: Dict[str, str],
) -> dict:
    """Compare the nodes version from a built Driver to
    the origins version from module source code when a second
    Driver can't be instantiated

    :nodes_version: mapping from node name to its origin function hash
    :origins_version: mapping from origin function name to origin function hash
    :node_to_origin: mapping from node name to origin name
    """
    node_added = []
    code_diff = []
    origins_removed = []

    # iterate over nodes from built Driver
    for node_name, node_version in nodes_version.items():
        origin_name = node_to_origin[node_name]
        origin_version = origins_version.get(origin_name)

        # if origin_version is None, the node is new
        if origin_version is None:
            node_added.append(node_name)
            continue

        # if node_version != origin_version, the current node's origin
        # function existed before, but was edited since origin_version
        if node_version != origin_version:
            code_diff.append(node_name)

    nodes_origin_version = set(nodes_version.values())
    # iterate over origin functions from source code
    for origin_name, origin_version in origins_version.items():
        # if origin_version not found in nodes_version, the
        # origin function is not used by the built Driver
        if origin_version not in nodes_origin_version:
            origins_removed.append(origin_name)

    return dict(
        nodes_added=node_added,
        code_diff=code_diff,
        origins_removed=origins_removed,
    )


def diff_versions(current_map: Dict[str, str], reference_map: Dict[str, str]) -> dict:
    """Generic diff of two {name: hash} mappings (can be node or origin name)

    :mapping_v1: mapping from node (or function) name to its function hash
    :mapping_v2: mapping from node (or function) name to its function hash
    """
    current_only, reference_only, edit = [], [], []

    for node_name, v1 in current_map.items():
        v2 = reference_map.get(node_name)
        if v2 is None:
            current_only.append(node_name)
            continue

        if v1 != v2:
            edit.append(node_name)

    for node_name, v2 in reference_map.items():
        v1 = current_map.get(node_name)
        if v1 is None:
            reference_only.append(node_name)
            continue

    return dict(current_only=current_only, reference_only=reference_only, edit=edit)


def _custom_diff_style(
    *,
    node,
    node_class,
    current_only: List[str],
    reference_only: List[str],
    edit: List[str],
):
    """Custom visualization style for the diff of 2 dataflows"""
    if node.name in current_only:
        style = ({"fillcolor": "greenyellow"}, node_class, "Current only")

    elif node.name in reference_only:
        style = ({"fillcolor": "tomato1"}, node_class, "Reference only")

    elif node.name in edit:
        style = ({"fillcolor": "yellow2"}, node_class, "Edited")

    else:
        style = ({}, node_class, None)

    return style


def visualize_diff(
    current_dr: driver.Driver,
    reference_dr: driver.Driver,
    current_only: List[str],
    reference_only: List[str],
    edit: List[str],
):
    """Visualize the diff of 2 dataflows.

    Uses the union of sets of nodes from driver 1 and driver 2.
    """
    import functools

    from hamilton import graph

    all_nodes = set(reference_dr.graph.get_nodes()).union(set(current_dr.graph.get_nodes()))

    diff_style = functools.partial(
        _custom_diff_style,
        current_only=current_only,
        reference_only=reference_only,
        edit=edit,
    )

    return graph.create_graphviz_graph(
        nodes=all_nodes,
        comment="Diff viz",
        graphviz_kwargs=dict(),
        custom_style_function=diff_style,
        node_modifiers=dict(),
        strictly_display_only_nodes_passed_in=True,
    )


# TODO refactor ContextLoader to a class
# TODO support loading from pyproject.toml
def load_context(file_path: Path) -> dict:
    if not file_path.exists():
        raise FileNotFoundError(f"`{file_path}` doesn't exist.")

    extension = file_path.suffix
    if extension == ".json":
        context = _read_json_context(file_path)
    elif extension == ".py":
        context = _read_py_context(file_path)
    else:
        raise ValueError(f"Received extension `{extension}` is unsupported.")

    context = _validate_context(context)
    return context


def _validate_context(context: dict) -> dict:
    if context[CONFIG_HEADER] is None:
        context[CONFIG_HEADER] = {}

    if context[FINAL_VARS_HEADER] is None:
        context[FINAL_VARS_HEADER] = []

    if context[INPUTS_HEADER] is None:
        context[INPUTS_HEADER] = {}

    if context[OVERRIDES_HEADER] is None:
        context[OVERRIDES_HEADER] = {}

    return context


def _read_json_context(file_path: Path) -> dict:
    """"""
    import json

    data = json.load(file_path.open())

    context = {}
    for k in [
        CONFIG_HEADER,
        FINAL_VARS_HEADER,
        INPUTS_HEADER,
        OVERRIDES_HEADER,
    ]:
        context[k] = data.get(k, None)

    return context


def _read_py_context(file_path: Path) -> dict:
    import importlib

    spec = importlib.util.spec_from_file_location("cli_config", file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    context = {}
    for k in [
        CONFIG_HEADER,
        FINAL_VARS_HEADER,
        INPUTS_HEADER,
        OVERRIDES_HEADER,
    ]:
        context[k] = getattr(module, k, None)

    return context



#!/usr/bin/env python
"""The setup script."""

import warnings

from setuptools import setup

# don't fail if there are problems with the readme (happens within circleci)
try:
    with open("README.md") as readme_file:
        readme = readme_file.read()
except Exception:
    warnings.warn("README.md not found")
    readme = None


def get_version():
    version_dict = {}
    with open("hamilton/version.py") as f:
        exec(f.read(), version_dict)
    return ".".join(map(str, version_dict["VERSION"]))


VERSION = get_version()


setup(
    version=VERSION,
    long_description=readme,
    long_description_content_type="text/markdown",
    zip_safe=False,
)



# code in this module should no depend on much
from typing import Any, Callable, List, Optional, Set, Tuple, Union


def convert_output_value(
    output_value: Union[str, Callable, Any], module_set: Set[str]
) -> Tuple[Optional[str], Optional[str]]:
    """Converts output values that one can request into strings.

    It checks that if it's a function, it's in the passed in module set.

    :param output_value: the value we want to convert into a string. We don't annotate driver.Variable here for
       import reasons.
    :param module_set: the set of modules functions could come from.
    :return: a tuple, (string value, string error). One or the other is returned, never both.
    """
    if isinstance(output_value, str):
        return output_value, None
    elif hasattr(output_value, "name"):
        return output_value.name, None
    elif isinstance(output_value, Callable):
        if output_value.__module__ in module_set:
            return output_value.__name__, None
        else:
            return None, (
                f"Function {output_value.__module__}.{output_value.__name__} is a function not "
                f"in a "
                f"module given to the materializer. Valid choices are {module_set}."
            )
    else:
        return None, (
            f"Materializer dependency {output_value} is not a string, a function, or a driver.Variable."
        )


def convert_output_values(
    output_values: List[Union[str, Callable, Any]], module_set: Set[str]
) -> List[str]:
    """Checks & converts outputs values to strings. This is used in building dependencies for the DAG.

    :param output_values: the values to convert.
    :param module_set: the modules any functions could come from.
    :return: the final values
    :raises ValueError: if there are values that can't be used/converted.
    """
    final_values = []
    errors = []
    for final_var in output_values:
        _val, _error = convert_output_value(final_var, module_set)
        if _val:
            final_values.append(_val)
        if _error:
            errors.append(_error)
    if errors:
        errors.sort()
        error_str = f"{len(errors)} errors encountered:\n  " + "\n  ".join(errors)
        raise ValueError(error_str)
    return final_values



from types import ModuleType
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import pytest

from hamilton import ad_hoc_utils, driver, node
from hamilton.io.materialization import to
from hamilton.lifecycle.base import (
    BaseDoNodeExecute,
    BasePostGraphConstruct,
    BasePostGraphExecute,
    BasePostNodeExecute,
    BasePreDoAnythingHook,
    BasePreGraphExecute,
    BasePreNodeExecute,
    ValidationException,
)
from hamilton.node import Node

from .lifecycle_adapters_for_testing import (
    ExtendToTrackCalls,
    SentinelException,
    TrackingDoBuildResultMethod,
    TrackingDoNodeExecuteHook,
    TrackingDoValidateInputMethod,
    TrackingPostGraphConstructHook,
    TrackingPostGraphExecuteHook,
    TrackingPostNodeExecuteHook,
    TrackingPreDoAnythingHook,
    TrackingPreGraphExecuteHook,
    TrackingPreNodeExecuteHook,
    TrackingValidateGraphValidator,
    TrackingValidateNodeValidator,
)

if TYPE_CHECKING:
    from hamilton.graph import FunctionGraph


def _sample_driver(*lifecycle_adapters):
    def a(input: int) -> int:
        return int(input) % 7  # Cause we're messing with the inputs for a lifecycle method

    def b(a: int) -> int:
        return a * 2

    def c(a: int, b: int, broken: bool = False) -> int:
        if broken:
            raise SentinelException("broken")
        return a * 3 + b * 2

    def d(c: int) -> int:
        return c**3

    mod = ad_hoc_utils.create_temporary_module(a, b, c, d)
    return driver.Builder().with_modules(mod).with_adapters(*lifecycle_adapters).build()


def test_individual_pre_do_anything_hook():
    hook_name = "pre_do_anything"
    hook = TrackingPreDoAnythingHook(name=hook_name)
    dr = _sample_driver(hook)
    dr.execute(["d"], inputs={"input": 1})
    relevant_calls = [item for item in hook.calls if item.name == hook_name]
    assert len(relevant_calls) == 1
    (adapter_call,) = relevant_calls
    assert adapter_call.bound_kwargs == {}


def test_individual_post_graph_construct_hook():
    hook_name = "post_graph_construct"
    hook = TrackingPostGraphConstructHook(name=hook_name)
    dr = _sample_driver(hook)
    dr.execute(["d"], inputs={"input": 1})
    relevant_calls = [item for item in hook.calls if item.name == hook_name]
    assert len(relevant_calls) == 1
    (adapter_call,) = relevant_calls
    assert adapter_call.fn.__name__ == hook.post_graph_construct.__name__
    assert "graph" in adapter_call.bound_kwargs
    assert "modules" in adapter_call.bound_kwargs
    assert "config" in adapter_call.bound_kwargs


def test_individual_pre_graph_execute_hook():
    hook_name = "pre_graph_execute"
    hook = TrackingPreGraphExecuteHook(name=hook_name)
    dr = _sample_driver(hook)
    dr.execute(["d"], inputs={"input": 1})
    relevant_calls = [item for item in hook.calls if item.name == hook_name]
    assert len(relevant_calls) == 1
    (adapter_call,) = relevant_calls
    assert adapter_call.fn.__name__ == "pre_graph_execute"
    assert "run_id" in adapter_call.bound_kwargs
    run_ids = {item.bound_kwargs["run_id"] for item in relevant_calls}
    assert len(run_ids) == 1
    (run_id,) = run_ids
    assert run_id is not None
    assert len(run_id) > 10  # Just a sanity check
    assert "graph" in adapter_call.bound_kwargs
    assert "final_vars" in adapter_call.bound_kwargs
    assert "inputs" in adapter_call.bound_kwargs
    assert "overrides" in adapter_call.bound_kwargs


def test_individual_pre_node_execute_hook():
    hook_name = "pre_node_execute"
    hook = TrackingPreNodeExecuteHook(name=hook_name)
    dr = _sample_driver(hook)
    dr.execute(["d"], inputs={"input": 1})
    relevant_calls = [item for item in hook.calls if item.name == hook_name]
    assert len(relevant_calls) == 4  # 4 nodes
    nodes_executed = {item.bound_kwargs["node_"].name for item in relevant_calls}
    assert nodes_executed == {"a", "b", "c", "d"}
    run_ids = {item.bound_kwargs["run_id"] for item in relevant_calls}
    assert len(run_ids) == 1
    (run_id,) = run_ids
    assert run_id is not None
    assert len(run_id) > 10  # Just a sanity check
    task_ids = {item.bound_kwargs["task_id"] for item in relevant_calls}
    assert len(task_ids) == 1
    (task_id,) = task_ids
    assert task_id is None


def test_individual_post_node_execute_hook():
    hook_name = "post_node_execute"
    hook = TrackingPostNodeExecuteHook(name=hook_name)
    dr = _sample_driver(hook)
    dr.execute(["d"], inputs={"input": 1})
    relevant_calls = [item for item in hook.calls if item.name == hook_name]
    assert len(relevant_calls) == 4  # 4 nodes
    nodes_executed = {item.bound_kwargs["node_"].name for item in relevant_calls}
    assert nodes_executed == {"a", "b", "c", "d"}
    assert {item.bound_kwargs["success"] for item in relevant_calls} == {True}
    assert {item.bound_kwargs["error"] for item in relevant_calls} == {None}
    task_ids = {item.bound_kwargs["task_id"] for item in relevant_calls}
    assert len(task_ids) == 1
    (task_id,) = task_ids
    assert task_id is None
    run_ids = {item.bound_kwargs["run_id"] for item in relevant_calls}
    assert len(run_ids) == 1
    (run_id,) = run_ids
    assert run_id is not None
    assert len(run_id) > 10  # Just a sanity check


def test_individual_post_node_execute_hook_with_exception():
    hook_name = "post_node_execute"
    hook = TrackingPostNodeExecuteHook(name=hook_name)
    dr = _sample_driver(hook)
    with pytest.raises(SentinelException):
        dr.execute(["d"], inputs={"input": 1, "broken": True})
    relevant_calls = [item for item in hook.calls if item.name == hook_name]
    assert len(relevant_calls) == 3  # 3 nodes ran, one failed
    nodes_executed = {item.bound_kwargs["node_"].name for item in relevant_calls}
    assert nodes_executed == {"a", "b", "c"}
    assert {item.bound_kwargs["success"] for item in relevant_calls} == {
        True,
        False,
    }  # 2 success, 1 failure
    errors = {item.bound_kwargs["error"] for item in relevant_calls}
    assert len(errors) == 2  # one error, one None
    task_ids = {item.bound_kwargs["task_id"] for item in relevant_calls}
    assert len(task_ids) == 1
    (task_id,) = task_ids
    assert task_id is None
    run_ids = {item.bound_kwargs["run_id"] for item in relevant_calls}
    assert len(run_ids) == 1
    (run_id,) = run_ids
    assert run_id is not None
    assert len(run_id) > 10  # Just a sanity check


def test_individual_post_graph_execute_hook():
    hook_name = "post_graph_execute"
    hook = TrackingPostGraphExecuteHook(name=hook_name)
    dr = _sample_driver(hook)
    dr.execute(["d"], inputs={"input": 1})
    relevant_calls = [item for item in hook.calls if item.name == hook_name]
    assert len(relevant_calls) == 1  # 1 task
    (adapter_call,) = relevant_calls
    assert adapter_call.fn.__name__ == "post_graph_execute"
    assert "run_id" in adapter_call.bound_kwargs
    assert "graph" in adapter_call.bound_kwargs
    assert adapter_call.bound_kwargs["success"] is True
    assert adapter_call.bound_kwargs["error"] is None
    assert adapter_call.bound_kwargs["results"] is not None
    run_ids = {item.bound_kwargs["run_id"] for item in relevant_calls}
    assert len(run_ids) == 1
    (run_id,) = run_ids
    assert run_id is not None
    assert len(run_id) > 10  # Just a sanity check


def test_individual_post_graph_execute_hook_with_exception():
    hook_name = "post_graph_execute"
    hook = TrackingPostGraphExecuteHook(name=hook_name)
    dr = _sample_driver(hook)
    with pytest.raises(SentinelException):
        dr.execute(["d"], inputs={"input": 1, "broken": True})
    relevant_calls = [item for item in hook.calls if item.name == hook_name]
    assert len(relevant_calls) == 1  # 1 task
    (adapter_call,) = relevant_calls
    assert adapter_call.fn.__name__ == "post_graph_execute"
    assert adapter_call.bound_kwargs["success"] is False
    assert isinstance(adapter_call.bound_kwargs["error"], SentinelException)
    assert adapter_call.bound_kwargs["results"] is None
    run_ids = {item.bound_kwargs["run_id"] for item in relevant_calls}
    assert len(run_ids) == 1
    (run_id,) = run_ids
    assert run_id is not None
    assert len(run_id) > 10  # Just a sanity check


def test_multiple_hooks():
    """Tests that all hooks are passed in"""
    all_hooks = [
        TrackingPreDoAnythingHook(name="pre_do_anything"),
        TrackingPostGraphConstructHook(name="post_graph_construct"),
        TrackingPreGraphExecuteHook(name="pre_graph_execute"),
        TrackingPreNodeExecuteHook(name="pre_node_execute"),
        TrackingPostNodeExecuteHook(name="post_node_execute"),
        TrackingPostGraphExecuteHook(name="post_graph_execute"),
    ]
    dr = _sample_driver(*all_hooks)
    dr.execute(["d"], inputs={"input": 1})
    # We want to make sure that the order of calls is correct
    calls_by_name = dict((hook.name, hook.calls) for hook in all_hooks)
    sequence_number_current = -1
    for hook_name in [
        "pre_do_anything",
        "post_graph_construct",
        "pre_graph_execute",
        "post_graph_execute",
    ]:
        assert len(calls_by_name[hook_name]) == 1
        (call,) = calls_by_name[hook_name]
        assert call.sequence_number > sequence_number_current
        sequence_number_current = call.sequence_number

    for hook_name in ["pre_node_execute", "post_node_execute"]:
        assert len(calls_by_name[hook_name]) == 4


def test_multiple_hooks_with_execution_exception():
    all_hooks = [
        TrackingPreDoAnythingHook(name="pre_do_anything"),
        TrackingPostGraphConstructHook(name="post_graph_construct"),
        TrackingPreGraphExecuteHook(name="pre_graph_execute"),
        TrackingPreNodeExecuteHook(name="pre_node_execute"),
        TrackingPostNodeExecuteHook(name="post_node_execute"),
        TrackingPostGraphExecuteHook(name="post_graph_execute"),
    ]
    dr = _sample_driver(*all_hooks)
    with pytest.raises(SentinelException):
        dr.execute(["d"], inputs={"input": 1, "broken": True})
    calls_by_name = dict((hook.name, hook.calls) for hook in all_hooks)
    sequence_number_current = -1
    for hook_name in [
        "pre_do_anything",
        "post_graph_construct",
        "pre_graph_execute",
        "post_graph_execute",
    ]:
        assert len(calls_by_name[hook_name]) == 1
        (call,) = calls_by_name[hook_name]
        assert call.sequence_number > sequence_number_current
        sequence_number_current = call.sequence_number
    for hook_name in ["pre_node_execute", "post_node_execute"]:
        assert len(calls_by_name[hook_name]) == 3  # Each is executed 3 times


def test_multi_hook():
    class MultiHook(
        BasePreDoAnythingHook,
        BasePostGraphConstruct,
        BasePreGraphExecute,
        BasePreNodeExecute,
        BaseDoNodeExecute,
        BasePostNodeExecute,
        BasePostGraphExecute,
        ExtendToTrackCalls,
    ):
        def do_node_execute(
            self,
            run_id: str,
            node_: node.Node,
            kwargs: Dict[str, Any],
            task_id: Optional[str] = None,
        ):
            return node_(**kwargs)

        def pre_do_anything(self):
            pass

        def post_graph_construct(
            self, graph: "FunctionGraph", modules: List[ModuleType], config: Dict[str, Any]
        ):
            pass

        def pre_graph_execute(
            self,
            run_id: str,
            graph: "FunctionGraph",
            final_vars: List[str],
            inputs: Dict[str, Any],
            overrides: Dict[str, Any],
        ):
            pass

        def pre_node_execute(
            self, run_id: str, node_: Node, kwargs: Dict[str, Any], task_id: Optional[str] = None
        ):
            pass

        def post_node_execute(
            self,
            run_id: str,
            node_: node.Node,
            kwargs: Dict[str, Any],
            success: bool,
            error: Optional[Exception],
            result: Optional[Any],
            task_id: Optional[str] = None,
        ):
            pass

        def post_graph_execute(
            self,
            run_id: str,
            graph: "FunctionGraph",
            success: bool,
            error: Optional[Exception],
            results: Optional[Dict[str, Any]],
        ):
            pass

    multi_hook = MultiHook(name="multi_hook")

    dr = _sample_driver(multi_hook)
    dr.execute(["d"], inputs={"input": 1})
    calls = multi_hook.calls
    assert len(calls) == 16


def test_individual_do_validate_input_method():
    method_name = "do_validate_input"
    method = TrackingDoValidateInputMethod(name=method_name, valid=True)
    dr = _sample_driver(method)
    dr.execute(["d"], inputs={"input": "1"})  # this is allowed by the method
    relevant_calls = [item for item in method.calls if item.name == method_name]
    (adapter_call,) = relevant_calls
    assert adapter_call.fn.__name__ == "do_validate_input"
    assert adapter_call.bound_kwargs["input_value"] == "1"
    assert adapter_call.result is True


def test_individual_do_validate_input_method_invalid():
    method_name = "do_validate_input"
    method = TrackingDoValidateInputMethod(name=method_name, valid=False)
    dr = _sample_driver(method)
    with pytest.raises(ValueError):
        dr.execute(["d"], inputs={"input": "1"})
    relevant_calls = [item for item in method.calls if item.name == method_name]
    (adapter_call,) = relevant_calls
    assert adapter_call.fn.__name__ == "do_validate_input"
    assert adapter_call.bound_kwargs["input_value"] == "1"
    assert adapter_call.result is False


def test_individual_do_node_execute_method():
    method_name = "do_node_execute"
    method = TrackingDoNodeExecuteHook(name=method_name, additional_value=1)
    dr = _sample_driver(method)
    res = dr.execute(["d"], inputs={"input": 1})
    relevant_calls = [item for item in method.calls if item.name == method_name]
    assert len(relevant_calls) == 4
    assert res == {"d": 17**3 + 1}  # adding one to each one


def test_individual_do_build_results_method():
    method_name = "do_build_result"
    method = TrackingDoBuildResultMethod(name=method_name, result=-1)
    dr = _sample_driver(method)
    res = dr.execute(["d"], inputs={"input": 1})
    relevant_calls = [item for item in method.calls if item.name == method_name]
    assert len(relevant_calls) == 1
    (adapter_call,) = relevant_calls
    assert res == -1
    assert adapter_call.bound_kwargs == {"outputs": {"d": 7**3}}


def test_multiple_hooks_materialize():
    # Just a single test as the codepaths are largely the same
    all_hooks = [
        TrackingPreDoAnythingHook(name="pre_do_anything"),
        TrackingPostGraphConstructHook(name="post_graph_construct"),
        TrackingPreGraphExecuteHook(name="pre_graph_execute"),
        TrackingPreNodeExecuteHook(name="pre_node_execute"),
        TrackingPostNodeExecuteHook(name="post_node_execute"),
        TrackingPostGraphExecuteHook(name="post_graph_execute"),
    ]
    dr = _sample_driver(*all_hooks)
    with pytest.raises(SentinelException):
        dr.materialize(to.memory(id="...", dependencies=["d"]), inputs={"input": 1, "broken": True})
    calls_by_name = dict((hook.name, hook.calls) for hook in all_hooks)
    sequence_number_current = -1
    for hook_name in [
        "pre_do_anything",
        "pre_graph_execute",
        "post_graph_execute",
    ]:
        assert len(calls_by_name[hook_name]) == 1
        (call,) = calls_by_name[hook_name]
        assert call.sequence_number > sequence_number_current
        sequence_number_current = call.sequence_number
    # This is called twice -- once when the graph is contructed in
    assert len(calls_by_name["post_graph_construct"]) == 2
    for hook_name in ["pre_node_execute", "post_node_execute"]:
        assert len(calls_by_name[hook_name]) == 3  # Each is executed 3 times


def test_individual_validate_node_validator_valid():
    validator_name = "validate_node"
    validator = TrackingValidateNodeValidator(name=validator_name, valid=True, message=None)
    _sample_driver(validator)  # Instantiating it just works
    relevant_calls = [item for item in validator.calls if item.name == validator_name]
    # TODO -- consider if we want to limit it to just the user-facing nodes
    assert len(relevant_calls) == 6  # one for each node then one for both inputs...


def test_individual_validate_node_validator_invalid():
    validator_name = "validate_node"
    validator = TrackingValidateNodeValidator(
        name=validator_name, valid=False, message="test message"
    )
    with pytest.raises(ValidationException):  # it groups and then gets called from that
        _sample_driver(validator)  # These are raised on instantiation
    relevant_calls = [item for item in validator.calls if item.name == validator_name]
    assert (
        len(relevant_calls) == 6
    )  # one for each node, we call everything and error out at the end
    assert all(item.result[0] is False for item in relevant_calls)


def test_individual_graph_validator_valid():
    validator_name = "validate_graph"
    validator = TrackingValidateGraphValidator(name=validator_name, valid=True, message=None)
    _sample_driver(validator)  # Instantiating it just works
    relevant_calls = [item for item in validator.calls if item.name == validator_name]
    assert len(relevant_calls) == 1


def test_individual_graph_validator_invalid():
    validator_name = "validate_graph"
    validator = TrackingValidateGraphValidator(
        name=validator_name, valid=False, message="test message"
    )
    with pytest.raises(ValidationException):  # it groups and then gets called from that
        _sample_driver(validator)  # These are raised on instantiation
    relevant_calls = [item for item in validator.calls if item.name == validator_name]
    assert (
        len(relevant_calls) == 1
    )  # one for each node, we call everything and error out at the end
    (relevant_call,) = relevant_calls
    assert relevant_call.result[0] is False



import logging

from hamilton.plugins.h_ray import RayGraphAdapter, RayTaskExecutor, RayWorkflowGraphAdapter

logger = logging.getLogger(__name__)
logger.warning(
    "Importing from this module is deprecated. We have moved these features out of experimental!"
    " Please use hamilton.plugins.h_ray instead."
)

__all__ = ["RayGraphAdapter", "RayWorkflowGraphAdapter", "RayTaskExecutor"]



import dataclasses
from os import PathLike
from typing import IO, Any, Collection, Dict, List, Optional, Tuple, Type, Union

try:
    from matplotlib.artist import Artist
    from matplotlib.figure import Figure
    from matplotlib.transforms import Bbox
except ImportError:
    raise NotImplementedError("Matplotlib is not installed.")

from hamilton import registry
from hamilton.io import utils
from hamilton.io.data_adapters import DataSaver


@dataclasses.dataclass
class MatplotlibWriter(DataSaver):
    """Write Matplotlib figure as static image format
    ref: https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure
    """

    path: Union[str, PathLike, IO]
    dpi: Optional[Union[float, str]] = None
    format: Optional[str] = None
    metadata: Optional[Dict] = None
    bbox_inches: Optional[Union[str, Bbox]] = None
    pad_inches: Optional[Union[float, str]] = None
    facecolor: Optional[Union[str, float, Tuple]] = None
    edgecolor: Optional[Union[str, float, Tuple]] = None
    backend: Optional[str] = None
    orientation: Optional[str] = None
    papertype: Optional[str] = None
    transparent: Optional[bool] = None
    bbox_extra_artists: Optional[List[Artist]] = None
    pil_kwargs: Optional[Dict] = None

    def _get_saving_kwargs(self) -> dict:
        kwargs = {}
        if self.format is not None:
            kwargs["format"] = self.format
        if self.metadata is not None:
            kwargs["metadata"] = self.metadata
        if self.bbox_inches is not None:
            kwargs["bbox_inches"] = self.bbox_inches
        if self.pad_inches is not None:
            kwargs["pad_inches"] = self.pad_inches
        if self.facecolor is not None:
            kwargs["facecolor"] = self.facecolor
        if self.edgecolor is not None:
            kwargs["edgecolor"] = self.edgecolor
        if self.backend is not None:
            kwargs["backend"] = self.backend
        if self.orientation is not None:
            kwargs["orientation"] = self.orientation
        if self.papertype is not None:
            kwargs["papertype"] = self.papertype
        if self.transparent is not None:
            kwargs["transparent"] = self.transparent
        if self.bbox_extra_artists is not None:
            kwargs["bbox_extra_artists"] = self.bbox_extra_artists
        if self.pil_kwargs is not None:
            kwargs["pil_kwargs"] = self.pil_kwargs

        return kwargs

    def save_data(self, data: Figure) -> Dict[str, Any]:
        data.savefig(fname=self.path, **self._get_saving_kwargs())
        # TODO make utils.get_file_metadata() safer for when self.path is IO type
        return utils.get_file_metadata(self.path)

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [Figure]

    @classmethod
    def name(cls) -> str:
        return "plt"


def register_data_savers():
    for saver in [
        MatplotlibWriter,
    ]:
        registry.register_adapter(saver)


register_data_savers()

COLUMN_FRIENDLY_DF_TYPE = False



"""Below is a file from `databackend` v0.0.2 found on GitHub
permalink: https://github.com/machow/databackend/blob/214832bab6990098d52a4a2fae935f5d9ae8dc3c/databackend/__init__.py
"""

# Copyright (c) 2024 databackend contributors (MIT License)
#
# See https://github.com/machow/databackend

import importlib
import sys
from abc import ABCMeta


def _load_class(mod_name: str, cls_name: str):
    mod = importlib.import_module(mod_name)
    return getattr(mod, cls_name)


class _AbstractBackendMeta(ABCMeta):
    def register_backend(cls, mod_name: str, cls_name: str):
        cls._backends.append((mod_name, cls_name))
        cls._abc_caches_clear()


class AbstractBackend(metaclass=_AbstractBackendMeta):
    @classmethod
    def __init_subclass__(cls):
        if not hasattr(cls, "_backends"):
            cls._backends = []

    @classmethod
    def __subclasshook__(cls, subclass):
        for mod_name, cls_name in cls._backends:
            if mod_name not in sys.modules:
                # module isn't loaded, so it can't be the subclass
                # we don't want to import the module to explicitly run the check
                # so skip here.
                continue
            else:
                parent_candidate = _load_class(mod_name, cls_name)
                if issubclass(subclass, parent_candidate):
                    return True

        return NotImplemented



import dataclasses
from typing import Any, Collection, Dict, Optional, Tuple, Type

from kedro.io import DataCatalog

from hamilton import registry
from hamilton.io.data_adapters import DataLoader, DataSaver


@dataclasses.dataclass
class KedroSaver(DataSaver):
    """Use Kedro DataCatalog and Dataset to save results
    ref: https://docs.kedro.org/en/stable/data/advanced_data_catalog_usage.html

    .. code-block:: python

        from kedro.framework.session import KedroSession

        with KedroSession.create() as session:
            context = session.load_context()
            catalog = context.catalog

        dr.materialize(
            to.kedro(
                id="my_dataset__kedro",
                dependencies=["my_dataset"],
                dataset_name="my_dataset",
                catalog=catalog,
            )
        )
    """

    dataset_name: str
    catalog: DataCatalog

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [Any]

    def save_data(self, data: Any) -> Dict[str, Any]:
        self.catalog.save(name=self.dataset_name, data=data)
        return dict(success=True)

    @classmethod
    def name(cls) -> str:
        return "kedro"


@dataclasses.dataclass
class KedroLoader(DataLoader):
    """Use Kedro DataCatalog and Dataset to load data
    ref: https://docs.kedro.org/en/stable/data/advanced_data_catalog_usage.html

    .. code-block:: python

        from kedro.framework.session import KedroSession

        with KedroSession.create() as session:
            context = session.load_context()
            catalog = context.catalog

        dr.materialize(
            from_.kedro(
                target="input_table",
                dataset_name="input_table",
                catalog=catalog
            )
        )
    """

    dataset_name: str
    catalog: DataCatalog
    version: Optional[str] = None

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [Any]

    def load_data(self, type_: Type) -> Tuple[Any, Dict[str, Any]]:
        data = self.catalog.load(name=self.dataset_name, version=self.version)
        metadata = dict(dataset_name=self.dataset_name, version=self.version)
        return data, metadata

    @classmethod
    def name(cls) -> str:
        return "kedro"


def register_data_loaders():
    for loader in [
        KedroSaver,
        KedroLoader,
    ]:
        registry.register_adapter(loader)


register_data_loaders()

COLUMN_FRIENDLY_DF_TYPE = False
